"""Helper Function to construct a SourceReader and StreamBuffer in one"""

import contextlib
from typing import Callable, Any, Optional

from stream2py import SourceReader, StreamBuffer
from stream2py.utility.typing_hints import ComparableType, OpenInstance


@contextlib.contextmanager
def mk_stream_buffer(
    read_stream: Callable[[OpenInstance], Optional[Any]],
    open_stream: Callable[[], OpenInstance] = None,
    close_stream: Callable[[OpenInstance], None] = None,
    info: Callable[[OpenInstance], dict] = None,
    key: Callable[[Any], ComparableType] = None,
    maxlen: Optional[int] = 100,
    sleep_time_on_read_none_seconds: float = None,
    auto_drop: bool = True,
    enumerate_data: bool = True,
):
    """Helper Function to construct a SourceReader and StreamBuffer in one

    ::

        from time import sleep
        from stream2py import mk_stream_buffer

        # download txt from: https://www.gutenberg.org/cache/epub/1524/pg1524.txt
        FILE = 'hamlet.txt'
        with mk_stream_buffer(
            read_stream=lambda open_inst: next(open_inst, None),
            open_stream=lambda: open(FILE, 'rt'),
            close_stream=lambda open_inst: open_inst.close(),
            key=lambda read_data: read_data[0],
            maxlen=None,
        ) as stream_buffer:
            buffer_reader = stream_buffer.mk_reader()
            sleep(1)
            # do stuff with buffer_reader
            print(buffer_reader.range(start=40, stop=50))

    :param read_stream: takes open instance and returns the next read item
    :param open_stream: open stream and return an open instance such as a file descriptor
    :param close_stream: take open instance and close the instance
    :param info: dict help in describing the instance
    :param key: sort key for read items
    :param maxlen: max read items to keep in buffer window
    :param sleep_time_on_read_none_seconds: time to wait if data is not ready to be read
    :param auto_drop: False is useful when data can be read independent of time and will quickly
        overflow the buffer maxlen before data is consumed. StreamBuffer.drop() must be called to
        continue reading after buffer has filled up. Default True.
    :param enumerate_data: wrap read data in an enumerated tuple. Default True.
    :return:
    """

    class SimplifiedSourceReader(SourceReader):
        open_instance: OpenInstance = None
        open_time = 0
        sleep_time_on_read_none_s = sleep_time_on_read_none_seconds
        read_idx = 0

        def open(self):
            self.read_idx = 0
            if open_stream is not None:
                self.open_instance = open_stream()
            self.open_time = self.get_timestamp()

        def read(self):
            read_data = read_stream(self.open_instance)
            if enumerate_data:
                read_data = (self.read_idx, read_data)
            self.read_idx += 1

            return read_data

        def close(self):
            if close_stream is not None:
                close_stream(self.open_instance)

        @property
        def info(self):
            if info is not None:
                return info(self.open_instance)
            return {
                'open_time': self.open_time,
                'open_instance': str(self.open_instance),
            }

        def key(self, data: Any):
            if key is not None:
                return key(data)
            return data

    stream_buffer = StreamBuffer(
        source_reader=SimplifiedSourceReader(),
        maxlen=maxlen,
        sleep_time_on_read_none_s=sleep_time_on_read_none_seconds,
        auto_drop=auto_drop,
    )

    try:
        stream_buffer.start()
        yield stream_buffer
    finally:
        stream_buffer.stop()


if __name__ == '__main__':
    counter = iter(range(10))
    with mk_stream_buffer(
        read_stream=lambda open_inst: next(counter),
        open_stream=lambda: print('open'),
        close_stream=lambda open_inst: print('close'),
        auto_drop=False,
        maxlen=2,
    ) as count_stream_buffer:
        count_reader = count_stream_buffer.mk_reader()
        print(f'start reading {count_reader.source_reader_info}')
        for i in count_reader:
            # do stuff with read data
            print(i)
            if count_stream_buffer.auto_drop is False:
                count_stream_buffer.drop()

        print('done reading')



import ctypes
from contextlib import suppress
from dataclasses import dataclass
from pprint import pprint
from typing import List, Optional, Callable

from snap7.common import check_error
from snap7.exceptions import Snap7Exception
from snap7.types import (
    S7DataItem,
    S7AreaDB,
    S7WLBit,
    S7WLReal,
    S7WLByte,
    S7WLWord,
    S7WLDWord,
)

from stream2py import SourceReader
import snap7


def get_byte(_bytearray, byte_index):
    """
    Get the boolean value from location in bytearray
    """
    return _bytearray[byte_index]


@dataclass
class PlcDataItem:

    area: int
    word_len: int
    db_number: int
    start: int
    amount: int

    key: str  # for multiple items read, represents key name in returned dictionary
    convert: snap7.util.get_bool or snap7.util.get_real or snap7.util.get_real or snap7.util.get_int or snap7.util.get_string or snap7.util.get_dword
    convert_args: Optional or None

    def __init__(
        self,
        area: int,
        word_len: int,
        db_number: int,
        start: int,
        amount: int,
        key: str,
        convert: Callable,
        convert_args: Optional or None = None,
    ):

        self.area = area
        self.word_len = int(word_len)
        self.db_number = int(db_number)
        self.start = eval(str(start))
        self.amount = int(amount)
        self.key = str(key)
        self.convert = convert
        self.convert_args = convert_args

        # allocate memory for data
        _size = 0
        if self.word_len in [S7WLBit, S7WLByte]:
            _size = 1
        elif self.word_len == S7WLWord:
            _size = 2
        elif self.word_len == S7WLDWord:
            _size = 4
        elif self.word_len == S7WLReal:
            _size = 8

        assert _size != 0, 'Unknown word len'

        print(f'PlcDataItem {self.key}: size = {_size}, amount = {self.amount}')

        try:
            self.buffer = ctypes.create_string_buffer(_size * self.amount)
        except Exception as ex:
            print(f'ERROR: Failed to allocate string buffer for item {self.key} : {ex}')
            return

        try:
            self.buffer = ctypes.cast(
                ctypes.pointer(self.buffer), ctypes.POINTER(ctypes.c_uint8)
            )
        except Exception as ex:
            return

    def get_item(self):
        return S7DataItem(
            Area=self.area,
            WordLen=self.word_len,
            DBNumber=self.db_number,
            Start=self.start,
            Amount=self.amount,
            pData=self.buffer,
        )

    def decode_item(self, item_read: S7DataItem):
        check_error(item_read.Result)

        if self.convert_args is not None:
            return self.convert(item_read.pData, *self.convert_args)
        else:
            return self.convert(item_read.pData, 0)


class PlcRawRead:
    def __init__(self, ip_address: str, *, rack: int, slot: int, tcp_port: int = 102):

        self._plc = snap7.client.Client()
        self._ip_address = ip_address
        self._rack = rack
        self._slot = slot
        self._tcp_port = tcp_port

        self.plc_info = {}

    def open(self):
        self._plc.connect(self._ip_address, self._rack, self._slot, self._tcp_port)
        return self._plc.get_connected()

    @classmethod
    def todict(cls, struct):
        return dict((field, getattr(struct, field)) for field, _ in struct._fields_)

    def get_info(self):
        if self._plc.get_connected():

            with suppress(Snap7Exception):
                self.plc_info.update(cpu_info=self.todict(self._plc.get_cpu_info()))

            with suppress(Snap7Exception):
                self.plc_info.update(cpu_state=self._plc.get_cpu_state())

            with suppress(Snap7Exception):
                self.plc_info.update(pdu_len=self._plc.get_pdu_length())

        return self.plc_info

    def close(self):
        if self._plc.get_connected():
            self._plc.disconnect()

        self._plc.destroy()

    def write_items(self, items: List[PlcDataItem]) -> bool:
        return False

    def read_items(self, items: List[PlcDataItem]) -> List[dict] or None:

        _items = (S7DataItem * len(items))()
        for _i in range(0, len(items)):
            _items[_i] = items[_i].get_item()

        result, items_read = self._plc.read_multi_vars(_items)
        if result:
            return None

        _bt = SourceReader.get_timestamp()
        result = []
        for _idx in range(0, len(items)):
            result.append(
                {
                    'key': items[_idx].key,
                    'value': items[_idx].decode_item(items_read[_idx]),
                    'bt': _bt,
                }
            )

        return result


read_items = [
    PlcDataItem(
        key='PLC Motor Status',
        area=S7AreaDB,
        word_len=S7WLBit,
        db_number=1,
        start=0 * 8 + 0,  # bit offset
        amount=1,
        convert=snap7.util.get_bool,
        convert_args=(0, 0),
    ),
    PlcDataItem(
        key='PLC LED Status',
        area=S7AreaDB,
        word_len=S7WLBit,
        db_number=1,
        start=0 * 8 + 1,  # bit offset
        amount=1,
        convert=snap7.util.get_bool,
        convert_args=(0, 0),
    ),
    # PlcDataItem(
    #     key='PLC Motor Speed',
    #     area=S7AreaDB,
    #     word_len=S7WLByte,
    #     db_number=1,
    #     start=1,
    #     amount=1,
    #     convert=get_byte,
    #     convert_args=(0,)),
    #
    # PlcDataItem(
    #     key='PLC LED Brightness',
    #     area=S7AreaDB,
    #     word_len=S7WLByte,
    #     db_number=1,
    #     start=2,
    #     amount=1,
    #     convert=get_byte,
    #     convert_args=(0,)),
    PlcDataItem(
        key='NetHAT Motor Speed',
        area=S7AreaDB,
        word_len=S7WLByte,
        db_number=1,
        start=3,
        amount=1,
        convert=get_byte,
        convert_args=(0,),
    ),
    PlcDataItem(
        key='NetHAT LED Brightness',
        area=S7AreaDB,
        word_len=S7WLByte,
        db_number=1,
        start=4,
        amount=1,
        convert=get_byte,
        convert_args=(0,),
    ),
]


if __name__ == '__main__':
    plcTest = PlcRawRead('192.168.0.19', rack=0, slot=0)
    plcTest.open()
    pprint(plcTest.get_info())

    while True:
        _d = plcTest.read_items(read_items)
        #
        #     [
        #     PlcDataItem(
        #         key='temperature',
        #         area=S7AreaDB,
        #         word_len=S7WLReal,
        #         db_number=3,
        #         start=2,
        #         amount=1,
        #         convert=snap7.util.get_real),
        #
        #     PlcDataItem(
        #         key='led1',
        #         area=S7AreaDB,
        #         word_len=S7WLBit,
        #         db_number=3,
        #         start=0 * 8 + 0,  # bit ofsset
        #         amount=1,
        #         convert=snap7.util.get_bool,
        #         convert_args=(0, 0)),
        #
        #     PlcDataItem(
        #         key='led2',
        #         area=S7AreaDB,
        #         word_len=S7WLBit,
        #         db_number=3,
        #         start=0 * 8 + 1,  # bit ofsset
        #         amount=1,
        #         convert=snap7.util.get_bool,
        #         convert_args=(0, 0)),
        # ])
        pprint(_d)
        print()

    plcTest.close()

    """
        Output example of get info:
        {'cpu_info': {'ASName': b'S71500/ET200MP station_1',
              'Copyright': b'Original Siemens Equipment',
              'ModuleName': b'PLC_1',
              'ModuleTypeName': b'CPU 1511C-1 PN',
              'SerialNumber': b'S V-L9AL98812019'},
     'cpu_state': 'S7CpuStatusRun',
     'pdu_len': 480}

     Output example of get Item

    [{'key': 'temperature', 'value': 10.0, 'ts': 1583086607352911}, {'key': 'led1', 'value': False, 'ts': 1583086607352923}, {'key': 'led2', 'value': False, 'ts': 1583086607352927}]

    """



"""
[taped](https://github.com/otosense/taped)) provides a `LiveWf`, but one would like to have more control over
 it's behavior. Namely:
* Getting an exception or sentinel value (of user's choice) when "some data might have been lost"
* repeats of `live_wf[:100]` give me 0-100, then 100-200, then 200-300...
    - this would be more clear to all users with a live_wf.next(100) method: Actually not: read is better.
* `live_wf[i:j]` is such that i and j are in samples since the process started, with customizable behavior when data
requested is no longer (past) or not yet (future) in buffer (behaviors such as returning exception, or sentinel,
or (in the case of future data), simply block (sleep) until data is available...


**Problem**

`PyAudioSourceReader` doesn't handle error flags and doesn't keep track of frame index out of the box.

_Doesn't handle errors_: Meaning if `PyAudio` catches an error (most likely one to encounter
being the "InputOverFlow" error, meaning the data wasn't read fast enough), tags it with a flag,
you'll see the corresponding `PyAudioSourceReader` item, which will be placed as is on the buffer,
but no special handling of these errors is done implicitly. **You** need to chose what to do with this.

_Doesn't keep track of frame index_: If you are dumping the `PyAudioSourceReader` items on a buffer,
then reading that buffer with some (or multiple) readers, that reader might want to know if it missed an item or not.
It can do so with some calculation involving the timestamp information, sample rate,
etc. but time and data rate are rarely multiple of each other, so some complexity is needed to resolve phase.
It would be more convenient to have a frame index (i.e. offset -- relative to the very first frame we read).

**Solution**

- `FillErrorWithZeroesMixin` is an element of a solution to the "handle errors" problem.
It solves the problem by filling missing data with zeros.

- `FrameIndexAsKeyMixin` is a solution for "frame index tracking" problem.
This solution assumes that frame index is more important than timestamp.
"""

import time
from collections import namedtuple

from stream2py.sources.audio import (
    PyAudioSourceReader,
    FillErrorWithZeroesMixin,
)
from stream2py import StreamBuffer, BufferReader


# TODO: Questions to consider:
#  Should we really fill with zeros, not None? Is parametrization worth it?
#  Consider this:
#    - no 0 in code, use class property instead (so semantic is clear, and possibilities of extension): Open closed principle
#    - possibility that -1 is a better fillval??
class FrameIndexAsKeyMixin(FillErrorWithZeroesMixin):
    """Must be used with FillErrorWithZeroesMixin for accurate frame count in case of errors"""

    _total_frame_index = 0

    def open(self):
        self._total_frame_index = 0  # ?? What about FrameIndexAsKeyMixin._total_frame_index? Needed? Safety measure?
        return super(FrameIndexAsKeyMixin, self).open()

    def data_to_append(self, timestamp, waveform, frame_count, time_info, status_flags):
        frame_index = self._total_frame_index
        self._total_frame_index += frame_count
        return (
            frame_index,
            timestamp,
            waveform,
            frame_count,
            time_info,
            status_flags,
        )


class PyAudioWithNamedTupleMixin:
    _data_to_append_namedtuple = namedtuple(
        typename='PyAudioSourceReaderData',
        field_names=['timestamp', 'bytes', 'frame_count', 'time_info', 'status_flags',],
    )

    def key(self, data):
        return data.timestamp

    def data_to_append(self, timestamp, waveform, frame_count, time_info, status_flags):
        return self._data_to_append_namedtuple(
            timestamp, waveform, frame_count, time_info, status_flags
        )


# TODO: Consider __init_subclass__ to check on proper subclassing (like no conflicts of Mixin method names)


class PyAudioWithZeroedErrorsSourceReader(
    FillErrorWithZeroesMixin, PyAudioSourceReader
):
    """PyAudioSourceReader with source errors resulting in (null) value filling"""

    pass


class PyAudioWithZeroedErrorsAndFrameIndexingSourceReader(
    FrameIndexAsKeyMixin, PyAudioSourceReader
):
    """PyAudioSourceReader with frame index tracking"""

    pass


def list_recording_device_index_names():
    """List (index, name) of available recording devices"""
    return sorted(
        (d['index'], d['name'])
        for d in PyAudioSourceReader.list_device_info()
        if d['maxInputChannels'] > 0
    )


def device_info_by_index(index):
    return next(
        d for d in PyAudioSourceReader.list_device_info() if d['index'] == index
    )


# Slicing ##############################################################################################################


class PyAudioStreamBuffer(StreamBuffer):
    _internal_reader: BufferReader = None  # reader persists here after source is stopped

    def __init__(
        self,
        input_device_index,
        sr=None,
        width=2,
        channels=None,
        frames_per_buffer=None,
        *,
        source_reader_class=PyAudioSourceReader,
    ):
        _info = device_info_by_index(input_device_index)
        sr = int(sr or _info['defaultSampleRate'])
        frames_per_buffer = int(frames_per_buffer or sr / 10)
        super(PyAudioStreamBuffer, self).__init__(
            source_reader=source_reader_class(
                input_device_index=input_device_index,
                rate=sr,
                width=width,
                channels=int(channels or _info['maxInputChannels']),
                frames_per_buffer=frames_per_buffer,
            ),
            maxlen=source_reader_class.audio_buffer_size_seconds_to_maxlen(
                buffer_size_seconds=60, rate=sr, frames_per_buffer=frames_per_buffer,
            ),
            auto_drop=False,
            sleep_time_on_read_none_s=0.1,
        )

    @property
    def reader(self):
        if self._internal_reader is None:
            self._internal_reader = self.mk_reader()
        return self._internal_reader

    def start(self):
        if self._internal_reader is not None:  # reset reader on start
            del self._internal_reader
            self._internal_reader = None
        return super(PyAudioStreamBuffer, self).start()

    def __getitem__(self, item):
        if not isinstance(item, slice):
            item = slice(item, item + 1)
        reader = self.reader
        return reader.range(item.start, item.stop, item.step, start_le=True)


class PyAudioStreamBufferWithByteSlice(PyAudioStreamBuffer):
    def __getitem__(self, item):
        # TODO: handle steps
        # TODO: robust way of defining how to slice items into subitems without making any assumptions
        if not isinstance(item, slice):
            item = slice(item, item + 1)
        items = super(PyAudioStreamBufferWithByteSlice, self).__getitem__(
            slice(item.start, item.stop)
        )
        i_bytes = b''.join(_i[1] for _i in items)

        first_item_bt = items[0][0]
        first_item_tt = items[1][0]
        subitem_index_per_time = len(items[0][1]) / (first_item_tt - first_item_bt)
        last_item_bt = items[-1][0]
        last_item_tt = last_item_bt + (len(items[-1][1]) / subitem_index_per_time)

        i_start = int((item.start - first_item_bt) * subitem_index_per_time)
        i_stop = int((item.stop - last_item_tt) * subitem_index_per_time)

        return i_bytes[i_start : i_stop : item.step]


if __name__ == '__main__':
    # TODO: Explain what we're testing here. What is expected to be observed?
    # TODO: If not too much overhead, assert what is expected.
    # TODO: If not too much overhead, include these asserts as actual tests (copying this code to a pytest module)

    from stream2py.sources.audio import find_a_default_input_device_index

    device = find_a_default_input_device_index()

    stream_buffer = PyAudioStreamBuffer(device, sr=44100, frames_per_buffer=4096)

    print(f'\nTest 1: PyAudioSourceReader ####################')
    with stream_buffer:
        time.sleep(2)
        bt = PyAudioSourceReader.get_timestamp()
        time.sleep(2)
        tt = PyAudioSourceReader.get_timestamp()
        time.sleep(2)
        items = stream_buffer[bt:tt]

    print(f'{len(items)=}, ({items[0][0]=}, {items[-1][0]=}), ({bt=}, {tt=})')

    print(f'{stream_buffer[(bt + tt) / 2][0][0]=}')

    print(f'\nTest 2: PyAudioStreamBufferWithByteSlice ######################')

    stream_buffer_with_byte_slice = PyAudioStreamBufferWithByteSlice(
        device, sr=44100, frames_per_buffer=4096,
    )

    with stream_buffer_with_byte_slice:
        time.sleep(2)
        bt = PyAudioSourceReader.get_timestamp()
        time.sleep(2)
        tt = PyAudioSourceReader.get_timestamp()
        time.sleep(2)
        items = stream_buffer_with_byte_slice[bt:tt]

    print(f'{len(items)=}, {items[0:10]=}, ({bt=}, {tt=})')

    bt = stream_buffer_with_byte_slice.reader.head(peek=True)[0]
    tt = stream_buffer_with_byte_slice.reader.tail(peek=True)[0]
    wf_bytes = stream_buffer_with_byte_slice[bt:tt]
    print(f'{len(wf_bytes)=}')

    print(f'\nTest 3: BufferReader Queries #################################')

    with stream_buffer:
        time.sleep(5)
        reader = stream_buffer.mk_reader()
        print(reader.source_reader_info)
        bt = PyAudioSourceReader.get_timestamp()
        for x in range(3, 0, -1):
            print(x)
            time.sleep(1)
        tt = PyAudioSourceReader.get_timestamp()
        bt_tt_data = reader.range(bt, tt, start_le=True)

    print(len(bt_tt_data), bt - bt_tt_data[0][0], tt - bt_tt_data[-1][0])

    buffer_count = 5

    # iterating from reader
    print('   ... iterating from reader')

    with stream_buffer:
        reader = stream_buffer.mk_reader()
        for i, data in enumerate(reader):
            print(i, data[0])
            if i >= buffer_count:
                break

    # using next() with work around

    print('   ... using next() with work around')

    with stream_buffer:
        reader = stream_buffer.mk_reader()

        def yield_from_reader(reader):
            yield from reader

        reader_gen = yield_from_reader(reader)

        for i in range(buffer_count):
            data = next(reader_gen)
            print(i, data[0])

    # using BufferReader.next()

    print('   ... using BufferReader.next()')

    with stream_buffer:
        reader = stream_buffer.mk_reader()

        i = 0
        while i < buffer_count:
            data = reader.next(ignore_no_item_found=True)
            if data:
                print(i, data[0])
                i += 1

    # using BufferReader.next(n=5)
    # TODO: Not working: Repair.
    # print("   ... using BufferReader.next(n=5)")
    #
    # with stream_buffer:
    #     reader = stream_buffer.mk_reader()
    #
    #     i = 0
    #     while i < buffer_count:
    #         data = reader.next(5, ignore_no_item_found=True)
    #         if data:
    #             print(i, data[0][0], len(data))
    #             i += 1
    #             time.sleep(1)

    # Frame Count Indexing

    print('\nTest 4: Frame Count Indexing ##########################')

    frame_index_stream_buffer = PyAudioStreamBuffer(
        device,
        sr=44100,
        frames_per_buffer=4096,
        source_reader_class=PyAudioWithZeroedErrorsAndFrameIndexingSourceReader,
    )

    start_index = 4095
    end_index = 20000

    with frame_index_stream_buffer:
        time.sleep(1)
        reader = frame_index_stream_buffer.mk_reader()
        for x in range(3, 0, -1):
            print(x)
            time.sleep(1)
        with reader._buffer.reader_lock() as _reader:
            _start = _reader.find_le(start_index)[0]
            print(_start)
        data_range = reader.range(start_index, end_index, start_le=True)
    len(data_range), data_range[0][0], data_range[-1][0]

    # iterating from reader
    print('   ... iterating from reader')

    with frame_index_stream_buffer:
        reader = frame_index_stream_buffer.mk_reader()
        for i, data in enumerate(reader):
            print(i, data[0])
            if i >= buffer_count:
                break

    # using next() with work around
    print('   ... using next() with work around')

    with frame_index_stream_buffer:
        reader = frame_index_stream_buffer.mk_reader()

        def yield_from_reader(reader):
            yield from reader

        reader_gen = yield_from_reader(reader)

        for i in range(buffer_count):
            data = next(reader_gen)
            print(i, data[0])

    # using BufferReader.next()
    print('   ... using BufferReader.next()')

    with frame_index_stream_buffer:
        reader = frame_index_stream_buffer.mk_reader()

        i = 0
        while i < buffer_count:
            data = reader.next(ignore_no_item_found=True)
            if data:
                print(i, data[0])
                i += 1



"""
.. graphviz::

    digraph {
        rankdir=LR;
        "SourceReader" [shape=cds]
        "StreamBuffer" [shape=box3d]
        "SourceReader" -> "StreamBuffer" -> "BufferReader1";
        "StreamBuffer" -> "BufferReader2";
        "StreamBuffer" -> "BufferReader3";
    }

The Rationale
-------------

Stream2py enables you to create a single asynchronous Source Reader and simply distribute
thread-safe readers to one or more consumers for the same data window. Or in other words, read and
work on data without worrying about missing data.

Normal way to process audio recording::

    import pyaudio

    CHUNK = 1024
    WIDTH = 2
    CHANNELS = 1
    RATE = 44100
    RECORD_SECONDS = 5

    p = pyaudio.PyAudio()

    stream = p.open(
        format=p.get_format_from_width(WIDTH),
        channels=CHANNELS,
        rate=RATE,
        input=True,
        frames_per_buffer=CHUNK
    )

    for i in range(0, int(RATE / CHUNK * RECORD_SECONDS)):
        data = stream.read(CHUNK)
        # risk of buffer overflow before next read
        long_calculations(data)

    stream.stop_stream()
    stream.close()
    p.terminate()

Stream2py way to process audio recording::

    from stream2py.sources.audio import PyAudioSourceReader

    CHUNK = 1024
    WIDTH = 2
    CHANNELS = 1
    RATE = 44100
    RECORD_SECONDS = 5

    source_reader = PyAudioSourceReader(
        rate=RATE,
        width=WIDTH,
        channels=CHANNELS,
        frames_per_buffer=CHUNK
    )
    with StreamBuffer(
        source_reader=source_reader, maxlen=int(RATE / CHUNK * RECORD_SECONDS)
    ) as stream_buffer:
        buffer_reader = stream_buffer.mk_reader()

        for i in range(0, int(RATE / CHUNK * RECORD_SECONDS)):
            data = next(buffer_reader)
            long_calculations(data)

How to use
----------
There are 3 components

1. The SourceReader
^^^^^^^^^^^^^^^^^^^
A SourceReader defines how to get data with the methods: open(), read(), and close(), and also how
the data is ordered with the key() method and an info property describing the instance.

::

    from typing import Tuple
    from stream2py import SourceReader

    EnumeratedReadData = Tuple[int, str]

    class LineSourceReader(SourceReader):
        def __init__(self, file, mode='rt'):
            self.file = file
            self.mode = mode
            self.line_count = 0
            self._f = None

        def open(self):
            self.line_count = 0
            self._f = open(self.file, self.mode)

        def read(self) -> EnumeratedReadData:
            read_data = (self.line_count, self._f.readline())
            self.line_count += 1
            return read_data

        def close(self):
            self._f.close()

        def key(self, data: EnumeratedReadData):
            return data[0]

        @property
        def info(self)
            return dict(file=self.file, mode=self.mode)

2. The StreamBuffer
^^^^^^^^^^^^^^^^^^^
A StreamBuffer has 2 jobs: First, it manages the open, read, and close of a SourceReader and puts
read data onto a thread-safe buffer.
Second, it is a factory of BufferReaders instances for multiple consumers.

::

    # ...Continued from above

    # download txt from: https://www.gutenberg.org/cache/epub/1524/pg1524.txt
    FILE = 'hamlet.txt'
    source_reader = LineSourceReader(FILE)

    with StreamBuffer(
        source_reader=source_reader,
        maxlen=None,  # None should only be used when reading objects with fixed lengths like a file
    ) as stream_buffer:
        buffer_reader = stream_buffer.mk_reader()

3. The BufferReader
^^^^^^^^^^^^^^^^^^^
A BufferReader gives data access to any number of consumers and provides methods to seek data such
as next(), range(), head(), tail().  Each BufferReader instance has it's own cursor keeping track of
what data was last seen.

::

    # ...Continued from above

        line_1 = buffer_reader.next()
        line_2 = next(buffer_reader)
        line_1_10 = buffer_reader.range(start=1, stop=10)

The All-in-One Helper
^^^^^^^^^^^^^^^^^^^^^
Here's a bonus helper function to shortcut the SourceReader class.

::

    from time import sleep
    from stream2py import mk_stream_buffer

    FILE = 'hamlet.txt'
    with mk_stream_buffer(
        read_stream=lambda open_inst: next(open_inst, None),
        open_stream=lambda: open(FILE, 'rt'),
        close_stream=lambda open_inst: open_inst.close(),
        key=lambda read_data: read_data[0],
        maxlen=None,
    ) as stream_buffer:
        buffer_reader = stream_buffer.mk_reader()
        sleep(1)
        # do stuff with buffer_reader
        print(buffer_reader.range(start=40, stop=50))

"""

from stream2py.source_reader import *
from stream2py.buffer_reader import *
from stream2py.stream_buffer import *
from stream2py.simply import mk_stream_buffer



import threading
import time
from asyncio import Queue
from collections import deque
from pprint import pprint
from typing import List, Optional, Any

import snap7
from snap7.types import S7AreaDB, S7WLReal, S7WLBit, S7WLByte

from stream2py import SourceReader
from stream2py.sources.raw_plc import PlcRawRead, PlcDataItem, get_byte

from stream2py.utility.typing_hints import ComparableType


class PlcReader(SourceReader):
    """
        TODO: Finish class implementation
    """

    def __init__(
        self,
        ip_address: str,
        *,
        items_to_read: List[PlcDataItem],
        rack: int,
        slot: int,
        tcp_port: int = 102,
        sleep_time=1.0
    ):

        self._init_kwargs = {
            k: v for k, v in locals().items() if k not in ('self', '__class__')
        }

        self._ip_address = ip_address
        self._rack = rack
        self._slot = slot
        self._tcp_port = tcp_port
        self._items_to_read = items_to_read
        self._sleep_time = sleep_time

        # validate IP address
        import socket

        socket.inet_aton(self._ip_address)  # validate IP Address
        self._plc_raw_reader = PlcRawRead(
            self._ip_address, rack=self._rack, slot=self._slot, tcp_port=self._tcp_port,
        )

        self.bt = None
        self._start_time = None
        # self._data_lock = threading.Lock()
        self._data_read_thread_exit = threading.Event()
        self._data_read_thread = None
        self._data_read_thread_exit.clear()
        self.data = deque()  # Queue()
        self.plc_info = dict()

        self.reader_thread = None

    def _stream_thread(self):
        _sleep_time = self.sleep_time_on_read_none_s

        while not self._data_read_thread_exit.is_set():
            data_item = self._plc_raw_reader.read_items(self._items_to_read)
            self.data.append(data_item)
            if _sleep_time > 0:
                time.sleep(_sleep_time)

    @property
    def sleep_time_on_read_none_s(self) -> float:
        return self._sleep_time

    def open(self) -> bool:

        if self._plc_raw_reader.open():
            self.bt = self.get_timestamp()
            self._start_time = self.bt
            self.plc_info = self._plc_raw_reader.get_info()
            if self._data_read_thread is None:
                self._data_read_thread = threading.Thread(target=self._stream_thread)
                self._data_read_thread.start()
            return True
        return False

    def read(self, blocking: bool = False, timeout: int = 0) -> Optional[Any]:
        """
        :return: timestamp, plc info, read db items as key:value
        """
        if len(self.data):
            return self.data.popleft()
        # return self.data.get(block = blocking, timeout=timeout)

    def close(self) -> None:
        """Close and clean up source reader.
        Will be called when StreamBuffer stops or if an exception is raised during read and append loop.
        """
        self._data_read_thread_exit.set()
        self._plc_raw_reader.close()

    @property
    def info(self) -> dict:
        _info = {'bt': self.bt}
        _info.update(**self._init_kwargs)
        _info.update(plc_info=self.plc_info)
        return _info

    def key(self, data_item: Any or None) -> ComparableType:

        assert (
            data_item is not None and len(data_item) > 0 and 'bt' in data_item[0]
        ), 'Cannot get key because bt is missing from data_item'

        import operator

        return operator.itemgetter('bt')(data_item[0])


if __name__ == '__main__':
    #
    # read_items = [
    #     PlcDataItem(
    #         key='temperature',
    #         area=S7AreaDB,
    #         word_len=S7WLReal,
    #         db_number=3,
    #         start=2,
    #         amount=1,
    #         convert=snap7.util.get_real),
    #
    #     PlcDataItem(
    #         key='led1',
    #         area=S7AreaDB,
    #         word_len=S7WLBit,
    #         db_number=3,
    #         start=0 * 8 + 0,  # bit ofsset
    #         amount=1,
    #         convert=snap7.util.get_bool,
    #         convert_args=(0, 0)),
    #
    #     PlcDataItem(
    #         key='led2',
    #         area=S7AreaDB,
    #         word_len=S7WLBit,
    #         db_number=3,
    #         start=0 * 8 + 1,  # bit ofsset
    #         amount=1,
    #         convert=snap7.util.get_bool,
    #         convert_args=(0, 0)),
    # ]

    read_items = [
        PlcDataItem(
            key='PLC Motor Status',
            area=S7AreaDB,
            word_len=S7WLBit,
            db_number=1,
            start=0 * 8 + 0,  # bit offset
            amount=1,
            convert=snap7.util.get_bool,
            convert_args=(0, 0),
        ),
        PlcDataItem(
            key='PLC LED Status',
            area=S7AreaDB,
            word_len=S7WLBit,
            db_number=1,
            start=0 * 8 + 1,  # bit offset
            amount=1,
            convert=snap7.util.get_bool,
            convert_args=(0, 0),
        ),
        PlcDataItem(
            key='NetHAT Motor Speed',
            area=S7AreaDB,
            word_len=S7WLByte,
            db_number=1,
            start=3,
            amount=1,
            convert=get_byte,
            convert_args=(0,),
        ),
        PlcDataItem(
            key='NetHAT LED Brightness',
            area=S7AreaDB,
            word_len=S7WLByte,
            db_number=1,
            start=4,
            amount=1,
            convert=get_byte,
            convert_args=(0,),
        ),
    ]

    preader = PlcReader(
        '192.168.0.19', items_to_read=read_items, rack=0, slot=0, sleep_time=0
    )

    if not preader.open():
        preader.close()
        exit(-1)
    can_run: bool = True

    _i = preader.info
    pprint(preader.info)
    while can_run:
        try:
            data = preader.read()
            if data is None:
                time.sleep(0.5)
            else:
                pass
                # pprint(data)
                # print()
        except KeyboardInterrupt as kb:
            can_run = False


"""

    Ouptut::

{'bt': 1584040986041418,
 'ip_address': '192.168.0.19',
 'items_to_read': [PlcDataItem(area=132, word_len=8, db_number=3, start=2, amount=1, key='temperature', convert=<function get_real at 0x7f8488038200>, convert_args=None),
                   PlcDataItem(area=132, word_len=1, db_number=3, start=0, amount=1, key='led1', convert=<function get_bool at 0x7f84780c2b00>, convert_args=(0, 0)),
                   PlcDataItem(area=132, word_len=1, db_number=3, start=1, amount=1, key='led2', convert=<function get_bool at 0x7f84780c2b00>, convert_args=(0, 0))],
 'plc_info': {'cpu_info': {'ASName': b'S71500/ET200MP station_1',
                           'Copyright': b'Original Siemens Equipment',
                           'ModuleName': b'PLC_1',
                           'ModuleTypeName': b'CPU 1511C-1 PN',
                           'SerialNumber': b'S V-L9AL98812019'},
              'cpu_state': 'S7CpuStatusRun',
              'pdu_len': 480},
 'rack': 0,
 'sleep_time': 1.0,
 'slot': 0,
 'tcp_port': 102}



[{'key': 'temperature', 'ts': 1584040986051531, 'value': 11.0},
 {'key': 'led1', 'ts': 1584040986051538, 'value': False},
 {'key': 'led2', 'ts': 1584040986051540, 'value': True}]

[{'key': 'temperature', 'ts': 1584040987061010, 'value': 11.0},
 {'key': 'led1', 'ts': 1584040987061022, 'value': False},
 {'key': 'led2', 'ts': 1584040987061029, 'value': True}]

[{'key': 'temperature', 'ts': 1584040988080404, 'value': 11.0},
 {'key': 'led1', 'ts': 1584040988080412, 'value': False},
 {'key': 'led2', 'ts': 1584040988080415, 'value': True}]

[{'key': 'temperature', 'ts': 1584040989104420, 'value': 11.0},
 {'key': 'led1', 'ts': 1584040989104428, 'value': False},
 {'key': 'led2', 'ts': 1584040989104431, 'value': True}]


"""



import logging
import os


def get_logging_level(env_variable_name, default=logging.NOTSET):
    if not isinstance(default, int):
        default = logging.NOTSET
    _value = os.getenv(env_variable_name, default)
    if _value == default:
        _level = _value
    else:
        _level = getattr(logging, _value, default)
        if not isinstance(_level, int):
            _level = default
    return _level


STREAM2PY_LOGGING_LEVEL = get_logging_level(
    env_variable_name='STREAM2PY_LOGGING_LEVEL', default=logging.WARNING
)


def set_logging_config(level=STREAM2PY_LOGGING_LEVEL, filename=None):
    logging.basicConfig(
        filename=filename,
        level=level,
        datefmt='%d-%m-%y %H:%M',
        format='%(asctime)s %(levelname)s %(filename)s:%(lineno)d  ::> %(message)s',
    )



from contextlib import contextmanager

from stream2py.utility.reader_writer_lock import RWLock
from stream2py.utility.sorted_deque import SortedDeque


class RWLockSortedDeque(RWLock):
    """
    >>> import operator
    >>> locked_deque = RWLockSortedDeque(
    ...     ((('plc', int(i // 10 * 10), i), f'data_{i}') for i in range(100)),
    ...     key=operator.itemgetter(0),
    ...     maxlen=100000)
    >>> new_item = (('plc', 100000, 100000), 'new data')
    >>> with locked_deque.writer_lock() as writer:
    ...     writer.append(new_item)
    >>> with locked_deque.reader_lock() as reader:
    ...     print(reader.find_gt(('plc', 90, 91)))
    ...
    (('plc', 90, 92), 'data_92')
    >>> with locked_deque.reader_lock() as reader:
    ...     print(reader.find_gt(('plc', 90, 99)))
    ...
    (('plc', 100000, 100000), 'new data')
    """

    def __init__(self, iterable=(), *, key=None, maxlen=None):
        RWLock.__init__(self)
        self._sorted_deque = SortedDeque(iterable, key, maxlen)

    def __len__(self):
        return len(self._sorted_deque)

    @property
    def key(self):
        return self._sorted_deque.key

    @contextmanager
    def reader_lock(self):
        try:
            self.reader_acquire()
            yield self._sorted_deque
        finally:
            self.reader_release()

    @contextmanager
    def writer_lock(self):
        try:
            self.writer_acquire()
            yield self._sorted_deque
        finally:
            self.writer_release()



__all__ = ['BufferReader']

import threading
import time
from contextlib import suppress

from stream2py.utility.locked_sorted_deque import RWLockSortedDeque
from stream2py.utility.typing_hints import Union


class BufferReader:
    """Reader that is constructed from StreamBuffer.mk_reader()

    >>> from stream2py import StreamBuffer, BufferReader
    >>> from stream2py.examples.source_reader import SimpleCounterString
    >>>
    >>> source_reader = SimpleCounterString(start=0, stop=100)
    >>> stream_buffer = StreamBuffer(source_reader=source_reader, maxlen=100)
    >>> stream_buffer.start()
    >>>
    >>> buffer_reader = stream_buffer.mk_reader()
    >>> isinstance(buffer_reader, BufferReader)
    True
    >>> buffer_reader.is_stopped
    False
    >>> buffer_reader.last_item == None  # None until value is seen
    True
    >>> buffer_reader.next() # get first item
    's0'
    >>> buffer_reader.last_item # check last_item cursor
    's0'
    >>> buffer_reader.next(peek=True), buffer_reader.last_item # next will not update last_item cursor
    ('s1', 's0')
    >>> buffer_reader.next(), buffer_reader.last_item # same as what was peeked but now cursor is updated
    ('s1', 's1')
    >>> buffer_reader.range(start=5, stop=10)
    ['s5', 's6', 's7', 's8', 's9', 's10']
    >>> buffer_reader.last_item
    's10'
    >>> buffer_reader.range(start=5, stop=10, step=3), buffer_reader.last_item
    (['s5', 's8'], 's8')
    >>> buffer_reader.range(start=10, stop=15, peek=True)
    ['s10', 's11', 's12', 's13', 's14', 's15']
    >>> buffer_reader.last_item # unchanged with peek
    's8'
    >>> buffer_reader.last_key # key of last_item
    8
    >>> buffer_reader.range(start=0, stop=9, only_new_items=True) # only 9 is greater than last item key
    ['s9']
    >>> buffer_reader.head(), buffer_reader.last_item, buffer_reader.last_key
    ('s0', 's0', 0)
    >>> buffer_reader.tail(), buffer_reader.last_item, buffer_reader.last_key
    ('s99', 's99', 99)
    >>>
    >>> stream_buffer.stop()
    >>>
    >>> buffer_reader.is_stopped
    True
    """

    def __init__(
        self,
        buffer: RWLockSortedDeque,
        source_reader_info: dict,
        stop_event: threading.Event,
    ):
        """

        :param buffer:
        :param source_reader_info:
        :param stop_event: threading.Event for source read loop
        """
        assert isinstance(buffer, RWLockSortedDeque)
        assert isinstance(stop_event, threading.Event)

        self._source_reader_info = source_reader_info
        self._buffer = buffer
        self._last_item = None
        self._last_key = None
        self._stop_event = stop_event
        self._sleep_time_on_iter_none_s = 0.1

    def __iter__(self):
        while True:
            _next = self.next(ignore_no_item_found=True)
            if _next is not None:
                yield _next
            elif self.is_stopped:
                return None
            else:
                time.sleep(self._sleep_time_on_iter_none_s)

    def __next__(self):
        return self.next(ignore_no_item_found=True)

    def set_sleep_time_on_iter_none(self, sleep_time_s: Union[int, float] = 0.1):
        """Set the sleep time of the iter yield loop when next data item is not yet available.

        :param sleep_time_s: seconds to sleep
        """
        self._sleep_time_on_iter_none_s = sleep_time_s

    def is_same_buffer(self, other_buffer_reader):
        """Check if reader is looking at the same buffer"""
        return (
            self.__class__ == other_buffer_reader.__class__
            and self._stop_event == other_buffer_reader._stop_event
        )

    @property
    def is_stopped(self) -> bool:
        """Checks if stop event has been set by StreamBuffer.

        :return: bool
        """
        return self._stop_event.is_set()

    @property
    def source_reader_info(self) -> dict:
        """A dict with important source info set by SourceReader.

        :return: dict
        """
        return self._source_reader_info

    @property
    def last_key(self):
        """key to last seen item cursor"""
        return self._last_key

    def _getlast_item(self):
        return self._last_item

    def _setlast_item(self, item):
        self._last_key = self._buffer.key(item)
        self._last_item = item

    def _dellast_item(self):
        del self._last_item
        self._last_item = None
        del self._last_key
        self._last_key = None

    last_item = property(
        _getlast_item, _setlast_item, _dellast_item, 'last seen item cursor'
    )

    def next(self, n=1, *, peek=False, ignore_no_item_found=False, strict_n=False):
        """Finds an item with a key greater than the last returned item.
        Raise ValueError if no item found with key above last item.

        :param n: number of items to return
        :param peek: if True, last_item cursor will not be updated
        :param ignore_no_item_found: if True, return None when no next item instead of raising exception
        :param strict_n: if True, raise ValueError if n items are not available
        :return: next item or list of next items if n > 1
        """
        with self._buffer.reader_lock() as reader:
            try:
                next_item = reader.find_gt(self.last_key)
            except ValueError as e:  # ValueError: No item found with key above: self.last_key
                if ignore_no_item_found:
                    return None
                else:
                    raise e
            except TypeError as e:  # TypeError: '<' not supported between instances of 'NoneType' and type(key)
                if self.last_item is None:  # first time reading a value from buffer
                    next_item = reader[0]
                else:
                    raise e
            if n > 1:
                i = reader.index(next_item)
                j = i + n
                if strict_n and j >= len(reader):
                    raise ValueError(
                        f'Number of items found is less than n: strict_n={strict_n}, n={n}'
                    )

                next_items_list = reader.range_by_index(i, j)
                if not peek:
                    self.last_item = next_items_list[-1]
                return next_items_list
            else:
                if not peek:
                    self.last_item = next_item
                return next_item

    def range(
        self,
        start,
        stop,
        step=None,
        *,
        peek=False,
        ignore_no_item_found=False,
        only_new_items=False,
        start_le=False,
        stop_ge=False,
    ):
        """Enables:
        1. Get last n minutes
        2. Give me data I don't have

        :param start: starting range key of item inclusive
        :param stop: ending range key of item inclusive
        :param step:
        :param peek: if True, last_item cursor will not be updated
        :param ignore_no_item_found: if True, return None when no next item instead of raising exception.
        :param only_new_items: if True and no new items, raise ValueError or return None if ignore_no_item_found
        :param start_le: if True, increase the search range to find start less than or equal by rounding down if start
            is in between keys, i.e. keys=[0, 10, 20], start=9 will include key=0
        :param stop_ge: if True, raise ValueError when there is no key greater than or equal to stop in buffer,
            if ignore_no_item_found is also True, return None instead of ValueError
        :return: list of items in range
        """
        with self._buffer.reader_lock() as reader:
            if only_new_items and self.last_key is not None:
                _next = self.next(peek=True, ignore_no_item_found=ignore_no_item_found)
                try:
                    _next_key = reader.key(_next)
                except TypeError as e:  # TypeError: 'NoneType' object is not subscriptable
                    if ignore_no_item_found:
                        return None
                    else:
                        raise e
                _start = start if start > _next_key else _next_key
            else:
                _start = start
            if start_le is True:
                with suppress(
                    ValueError
                ):  # ValueError: No item found with key at or below: _start
                    _start = reader.key(reader.find_le(_start))
            if stop_ge is True:
                try:
                    stop = reader.key(reader.find_ge(stop))
                except ValueError as e:  # ValueError: No item found with key at or above: stop
                    if ignore_no_item_found:
                        return None
                    else:
                        raise e

            items = reader.range(_start, stop, step)

        if not peek:
            try:
                self.last_item = items[-1]
            except IndexError as e:  # IndexError: list index out of range
                if ignore_no_item_found:
                    return None
                else:
                    raise e
        return items

    def tail(self, *, peek=False, ignore_no_item_found=False, only_new_items=False):
        """Finds the last item in buffer. Raise ValueError if no item found.

        :param peek: if True, last_item cursor will not be updated
        :param ignore_no_item_found: if True, return None when no next item instead of raising exception
        :param only_new_items: if True and no new items, raise ValueError or return None if ignore_no_item_found
        :return: tail item
        """
        with self._buffer.reader_lock() as reader:
            if only_new_items:
                try:
                    item = reader.find_last_gt(self.last_key)
                except ValueError as e:  # ValueError: No item found with key above: self.last_key
                    if ignore_no_item_found:
                        return None
                    else:
                        raise e
                except TypeError as e:  # TypeError: '>' not supported between instances of type(key) and 'NoneType'
                    if self.last_item is None:  # first time reading a value from buffer
                        item = reader[-1]
                    else:
                        raise e
            else:
                try:
                    item = reader[-1]
                except IndexError as e:  # IndexError: deque index out of range
                    if ignore_no_item_found:
                        return None
                    else:
                        raise e
        if not peek:
            self.last_item = item
        return item

    def head(self, *, peek=False):
        with self._buffer.reader_lock() as reader:
            item = reader[0]
        if not peek:
            self.last_item = item
        return item



__all__ = ['getch']


class _Getch:
    """
    Gets a single character from standard input.  Does not echo to the screen.
    """

    def __init__(self, is_blocking=True):
        try:
            self.impl = _GetchWindows(is_blocking)
        except (ImportError, ModuleNotFoundError):
            self.impl = _GetchUnix(is_blocking)

    def __getattr__(self, attr):
        return getattr(self.impl, attr)

    def __call__(self):
        return self.impl()


class _GetchUnix:
    def __init__(self, is_blocking):
        import tty, sys, termios

        self.fd = sys.stdin.fileno()
        self.old_settings = termios.tcgetattr(self.fd)
        if is_blocking is True:
            self._getch = self.blocking
        else:
            self._getch = self.non_blocking

    def blocking(self):
        import sys, tty, termios

        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch

    def non_blocking(self):
        import sys, tty, termios

        old_settings = termios.tcgetattr(sys.stdin)
        ch = None
        try:
            tty.setcbreak(sys.stdin.fileno())
            if self._is_data():
                ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_settings)
        return ch

    def restore_settings(self):
        import sys, tty, termios

        termios.tcsetattr(sys.stdin, termios.TCSADRAIN, self.old_settings)

    @staticmethod
    def _is_data():
        import select, sys

        return select.select([sys.stdin], [], [], 0) == ([sys.stdin], [], [])

    def __call__(self):
        return self._getch()


class _GetchWindows:
    def __init__(self, is_blocking):
        import msvcrt

        if is_blocking is True:
            self._getch = self.blocking
        else:
            self._getch = self.non_blocking

    def blocking(self):
        import msvcrt

        return msvcrt.getch()

    def non_blocking(self):
        import msvcrt

        if msvcrt.kbhit():
            return msvcrt.getch()

    def restore_settings(self):
        pass

    def __call__(self):
        return self._getch()


getch = _Getch()


if __name__ == '__main__':
    import sys

    def getch_loop(is_blocking=True):
        print(
            f'{"Blocking" if is_blocking is True else "Non-blocking"} getch! Press any key! Esc to quit!'
        )
        i = 0
        getch_func = getch.blocking if is_blocking is True else getch.non_blocking
        while True:
            char = getch_func()
            if char or i % 15000 == 0:
                print(f'{i}: {char}')

            if char == '\x1b':  # ESC key
                break
            i += 1

    getch_file, *args = sys.argv
    print(
        'Getch! Echo key press usage:\n'
        f'Blocking mode: python {getch_file}\n'
        f'Non-blocking mode: python {getch_file} False\n'
    )

    getch_loop(is_blocking=False if len(args) and args[0] == 'False' else True)



from collections import deque
from contextlib import suppress
import operator
import threading
import time
from typing import Any

from stream2py import SourceReader
from stream2py.utility.typing_hints import ComparableType


_ITEMGETTER_0 = operator.itemgetter(0)


class KeyboardInputSourceReader(SourceReader, threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self, daemon=True)
        self._data_lock = threading.Lock()
        self.data = deque()
        self.stop_event = threading.Event()
        self.bt = None
        self.index = 0

    def open(self):
        self.data.clear()
        self.stop_event.clear()
        self.bt = self.get_timestamp()
        self.index = 0
        self.start()

    def read(self):
        """Returns one data item

        :return: (index, timestamp, character)
        """
        if len(self.data):
            with self._data_lock:
                return self.data.popleft()

    def close(self):
        # Import getch here to avoid errors for rst preview extension when loaded within IDE terminals
        from stream2py.utility.getch import getch

        self.stop_event.set()
        getch.restore_settings()

    @property
    def info(self) -> dict:
        return {'bt': self.bt}

    def key(self, data: Any) -> ComparableType:
        """

        :param data: (index, timestamp, character)
        :return: index
        """
        return _ITEMGETTER_0(data)

    def run(self):
        # Import getch here to avoid errors for rst preview extension when loaded within IDE terminals
        from stream2py.utility.getch import getch

        try:
            while not self.stop_event.is_set():
                ch = getch.blocking()
                self.data.append(
                    (self.index, self.get_timestamp(), ch)
                )  # (index, timestamp, character)
                self.index += 1
        except Exception:
            self.close()
            raise


if __name__ == '__main__':
    with KeyboardInputSourceReader() as source:
        print('getch! Press any key! Esc to quit!')
        while True:
            data = source.read()
            if data is not None:
                index, timestamp, char = data
                print(f'{timestamp}: {char}', end='\n\r')

                if char == '\x1b':  # ESC key
                    break



"""
TODO:
    Maybe for next more advanced version:
    * default output files to ~/odir
    * set StreamBuffer maxlen to something reasonable
    * set VideoShow intervals based on frame_rate
    * add a lookup for VideoCapture info like frame_size and frame_rate for any give input
    * playback/recording controls
        [x] default just plays and records everything
        [ ] frame rate control
        [ ] skip back/ahead
        [ ] start/stop record based on what is currently displayed
            * need to sync VideoWriter and VideoShow
    What interface do I need for VideoWriter and VideoShow
    see how to handle syncing buffer_readers between two consumers
      sharing one buffer_reader won't work like anything expected or intended or desired
"""


from stream2py import StreamBuffer
from stream2py.sources.video import VideoCapture
from stream2py.sources.keyboard_input import KeyboardInputSourceReader
from stream2py.consumers.video.write import VideoWriter
from stream2py.consumers.video.show import VideoShow


def video_display_and_save(video_input, file_name='VideoWriter.avi', fourcc='MJPG'):
    """Display recording on screen and save to file

    TODO: fps info from devices (camera) is not accurate and need to be calculated

    :param video_input: file name or device id of video source
    :param file_name: video file path to save recording
    :param fourcc: video file encoding
    """
    source_reader = VideoCapture(video_input=video_input)
    with StreamBuffer(source_reader=source_reader, maxlen=1000) as stream_buffer:

        buffer_reader_for_write = stream_buffer.mk_reader()
        frame_size = (
            buffer_reader_for_write.source_reader_info['frame_width'],
            buffer_reader_for_write.source_reader_info['frame_height'],
        )
        fps = buffer_reader_for_write.source_reader_info['fps']

        with VideoWriter(
            buffer_reader=buffer_reader_for_write,
            interval=0.001,
            file_name=file_name,
            fourcc=fourcc,
            fps=fps,
            frame_size=frame_size,
        ) as writer:

            buffer_reader_for_show = stream_buffer.mk_reader()

            with VideoShow(
                buffer_reader=buffer_reader_for_show,
                interval=0.001,
                window_name='show_demo',
            ) as show:
                print('press "q" to end recording')
                with KeyboardInputSourceReader() as key_input:
                    for index, timestamp, char in iter(key_input):
                        if char == 'q':
                            print('quitting')
                            break


if __name__ == '__main__':
    video_display_and_save(video_input=0)



"""Example of processing audio and keyboard streams"""
from typing import Callable, NewType, Any
from stream2py.stream_buffer import StreamBuffer
from stream2py.sources.keyboard_input import KeyboardInputSourceReader
from stream2py.sources.audio import (
    PyAudioSourceReader,
    find_a_default_input_device_index,
)


AudioData = NewType('', Any)
KeyboardData = NewType('', Any)
AudioDataCallback = Callable[[AudioData], Any]
KeyboardDataCallback = Callable[[KeyboardData], Any]


def default_audio_callback(audio_data):
    if audio_data is not None:
        (audio_timestamp, waveform, frame_count, time_info, status_flags,) = audio_data
        print(
            f'   [Audio] {audio_timestamp}: {len(waveform)=} {type(waveform).__name__}',
            end='\n\r',
        )


def default_keyboard_event_callback(keyboard_data):
    if keyboard_data is not None:
        index, keyboard_timestamp, char = keyboard_data
        print(f'[Keyboard] {keyboard_timestamp}: {char}', end='\n\r')

        if char == '\x1b':  # ESC key
            return True
        else:
            return False


def keyboard_and_audio(
    input_device_index=None,  # find index with PyAudioSourceReader.list_device_info()
    rate=44100,
    width=2,
    channels=1,
    frames_per_buffer=44100,  # same as sample rate for 1 second intervals
    seconds_to_keep_in_stream_buffer=60,
    audio_data_callback: AudioDataCallback = default_audio_callback,
    keyboard_data_callback: KeyboardDataCallback = default_keyboard_event_callback,
):
    """Starts two independent streams: one for audio and another for keyboard inputs.
    Prints stream type, timestamp, and additional info about data:
    Shows input key pressed for keyboard and byte count for audio

    Press Esc key to quit.

    :param input_device_index: find index with PyAudioSourceReader.list_device_info()
    :param rate: audio sample rate
    :param width: audio byte width
    :param channels: number of audio input channels
    :param frames_per_buffer: audio samples per buffer
    :param seconds_to_keep_in_stream_buffer: max size of audio buffer before data falls off
    :return: None
    """

    if input_device_index is None:
        input_device_index = find_a_default_input_device_index()

    selected_device_info = next(
        dev
        for dev in PyAudioSourceReader.list_device_info()
        if dev['index'] == input_device_index
    )
    print(f"Starting audio device: {selected_device_info['name']}\n")

    # converts seconds_to_keep_in_stream_buffer to max number of buffers of size frames_per_buffer
    maxlen = PyAudioSourceReader.audio_buffer_size_seconds_to_maxlen(
        buffer_size_seconds=seconds_to_keep_in_stream_buffer,
        rate=rate,
        frames_per_buffer=frames_per_buffer,
    )
    audio_source_reader = PyAudioSourceReader(
        rate=rate,
        width=width,
        channels=channels,
        unsigned=True,
        input_device_index=input_device_index,
        frames_per_buffer=frames_per_buffer,
    )

    # open audio stream
    with StreamBuffer(
        source_reader=audio_source_reader, maxlen=maxlen
    ) as audio_stream_buffer:
        # open keyboard input stream
        with StreamBuffer(
            source_reader=KeyboardInputSourceReader(), maxlen=maxlen
        ) as keyboard_stream_buffer:

            audio_buffer_reader = audio_stream_buffer.mk_reader()
            keyboard_buffer_reader = keyboard_stream_buffer.mk_reader()

            print('getch! Press any key! Esc to quit!\n')  # replace w KeyboardInterrupt
            while True:
                try:
                    keyboard_data = next(keyboard_buffer_reader)
                    audio_data = next(audio_buffer_reader)

                    should_quit = keyboard_data_callback(keyboard_data)
                    if should_quit:
                        break

                    audio_data_callback(audio_data)

                except KeyboardInterrupt as e:
                    print(f"\n\nGot a {e}: I'll quit now...\n")
                    break


if __name__ == '__main__':
    keyboard_and_audio()



"""
Work in progress. Use as example to construct your own until this is finalized with a standard api.

Abstract classes to make asynchronous consumers of BufferReaders and StreamBuffers.
"""

from abc import ABCMeta, abstractmethod

from contextlib import suppress
import logging
import threading
import time

from stream2py import BufferReader, StreamBuffer
from stream2py.utility.typing_hints import Union

logger = logging.getLogger(__name__)


class BufferReaderConsumer(threading.Thread, metaclass=ABCMeta):
    """Call reader_handler function with reader at defined time intervals between calls"""

    def __init__(
        self,
        buffer_reader: BufferReader,
        interval: Union[int, float],
        logging_enabled: bool = False,
    ):
        """

        :param buffer_reader: BufferReader created from a StreamBuffer
        :param interval: seconds between reader_handler calls
        :param logging_enabled: log debug messages using logging module
        """
        threading.Thread.__init__(self, daemon=True)
        assert isinstance(buffer_reader, BufferReader)
        self.buffer_reader = buffer_reader
        self.interval = interval
        self.stop_event = threading.Event()
        self.logging_enabled = logging_enabled

    def stop(self):
        """Sets stop event and waits for thread to finish"""
        self.stop_event.set()
        if self.logging_enabled:
            logger.debug(f'Consumer stopping... {self.__class__.__name__}')
        self.join()

    def run(self):
        """Calls self.reader_handler in a loop while sleeping for self.interval seconds after each call until stop event
        is set or Exception is raised by self.reader_handler"""
        if self.logging_enabled:
            logger.debug(f'Consumer starting! {self.__class__.__name__}')
        try:
            while not self.stop_event.is_set():
                self.reader_handler(self.buffer_reader)
                if self.interval > 0:
                    time.sleep(self.interval)
        finally:
            with suppress(Exception):
                self.stop()
        if self.logging_enabled:
            logger.debug(f'Consumer stopped! {self.__class__.__name__}')

    @abstractmethod
    def reader_handler(self, buffer_reader: BufferReader):
        """Implements how to read from buffer and what to do with the data"""
        raise NotImplementedError

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


class StreamBufferConsumer(threading.Thread, metaclass=ABCMeta):
    def __init__(
        self,
        stream_buffer: StreamBuffer,
        interval: Union[int, float],
        logging_enabled: bool = False,
    ):
        """

        :param stream_buffer: a StreamBuffer
        :param interval: seconds between stream_buffer_handler calls
        :param logging_enabled: log debug messages using logging module
        """
        threading.Thread.__init__(self, daemon=True)
        assert isinstance(stream_buffer, StreamBuffer)
        self.stream_buffer = stream_buffer
        self.interval = interval
        self.stop_event = threading.Event()
        self.logging_enabled = logging_enabled

    def stop(self):
        """Sets stop event and waits for thread to finish"""
        self.stop_event.set()
        if self.logging_enabled:
            logger.debug(f'Consumer stopping... {self.__class__.__name__}')
        self.join()

    def run(self):
        """Calls self.reader_handler in a loop while sleeping for self.interval seconds after each call until stop event
        is set or Exception is raised by self.stream_buffer_handler"""
        if self.logging_enabled:
            logger.debug(f'Consumer starting! {self.__class__.__name__}')
        try:
            while not self.stop_event.is_set():
                self.stream_buffer_handler(self.stream_buffer)
                time.sleep(self.interval)
        finally:
            with suppress(Exception):
                self.stop()
        if self.logging_enabled:
            logger.debug(f'Consumer stopped! {self.__class__.__name__}')

    @abstractmethod
    def stream_buffer_handler(self, stream_buffer: StreamBuffer):
        """Implements how to read from buffer and what to do with the data"""
        raise NotImplementedError

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



from typing import Any, Optional, Union

import cv2
import operator

from stream2py import SourceReader
from stream2py.utility.typing_hints import ComparableType

_ITEMGETTER_0 = operator.itemgetter(0)


class VideoCapture(SourceReader):
    """Video Capture using OpenCV"""

    def __init__(self, video_input: Union[str, int] = 0):
        """
        https://docs.opencv.org/4.2.0/d8/dfe/classcv_1_1VideoCapture.html#ac4107fb146a762454a8a87715d9b7c96
        https://docs.opencv.org/4.2.0/d8/dfe/classcv_1_1VideoCapture.html#aabce0d83aa0da9af802455e8cf5fd181

        :param video_input: filename or device id, see cv2.VideoCapture documentation for more info
        """
        if isinstance(video_input, str):
            self.is_file = True
        elif isinstance(video_input, int):
            self.is_file = False
        else:
            raise TypeError(
                f'{self.__class__.__name__} video_input must be type str or int: {type(video_input)}'
            )
        self.video_capture = None
        self._bt = -1
        self.video_input = video_input

    @classmethod
    def video_input_info(cls, video_input: Union[str, int]) -> dict:
        """
        https://docs.opencv.org/4.2.0/d4/d15/group__videoio__flags__base.html#gaeb8dd9c89c10a5c63c139bf7c4f5704d

        :param video_input: filename or device id, see cv2.VideoCapture documentation for more info
        :return: dict

        .. todo:: filter for useful info
        """
        vid_cap = cv2.VideoCapture(video_input)
        _info = {
            cap_prop: vid_cap.get(getattr(cv2, cap_prop))
            for cap_prop in dir(cv2)
            if cap_prop.startswith('CAP_PROP_')
        }
        return _info

    def open(self) -> None:
        self._bt = self.get_timestamp()
        self.video_capture = cv2.VideoCapture(self.video_input)
        if self.is_opened() is False:
            raise IOError(
                f'{self.__class__.__name__} error opening video stream or file: {self.video_input}'
            )

    def read(self) -> Optional[Any]:
        ret, frame = self.video_capture.read()
        return self.get_timestamp(), ret, frame

    def close(self) -> None:
        self.video_capture.release()

    @property
    def info(self) -> dict:
        _info = {'video_input': self.video_input, 'bt': self._bt}
        if self.is_opened():
            _info.update(
                frame_width=int(self.video_capture.get(cv2.CAP_PROP_FRAME_WIDTH)),
                frame_height=int(self.video_capture.get(cv2.CAP_PROP_FRAME_HEIGHT)),
                fps=self.video_capture.get(cv2.CAP_PROP_FPS),
            )
        return _info

    def key(self, data) -> ComparableType:
        """
        :param data: (timestamp, ret, frame)
        :return: timestamp
        """
        return _ITEMGETTER_0(data)

    def is_opened(self) -> bool:
        return self.video_capture is not None and self.video_capture.isOpened()


if __name__ == '__main__':
    video_input = 0
    from pprint import pprint

    pprint(VideoCapture.video_input_info(video_input))

    print('starting recording')
    with VideoCapture(video_input=video_input) as cap:
        print(cap.info)
        print("Press 'q' to quit")
        while cap.is_opened():
            ts, ret, frame = cap.read()
            cv2.imshow('frame', frame)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break



#!/usr/bin/env python
import operator
import os
import threading
import time
from collections import deque
from pprint import pprint

import psutil
import platform

from stream2py import SourceReader
from stream2py.utility.typing_hints import ComparableType, Any

"""
    Provide computer status info. 
    
    To get network upload and download speeds, speedtest-cli app should be installed 
"""

__all__ = ['StatusInfo', 'StatusInfoReader']

DFLT_STATUS_INFO_READ_INTERVAL = 1000  # in ms


class StatusInfo:

    SPEEDTEST_CMD = 'speedtest'

    @staticmethod
    def disk_free_bytes(path: str = '/') -> dict:
        return {'val': int(psutil.disk_usage(path).free), 'unit': 'bytes'}

    @staticmethod
    def disk_used_percents(path: str = '/') -> dict:
        return {'val': float(psutil.disk_usage(path).percent), 'unit': '%'}

    @staticmethod
    def cpu_used_percents():
        return {'val': psutil.cpu_percent(), 'unit': '%'}

    @staticmethod
    def cpu_temp():

        _t = -1
        try:
            with open('/sys/class/thermal/thermal_zone0/temp', 'r') as ftemp:
                _t = ftemp.readline()
                _t = float(int(_t) / 1000)
        except Exception as ex:
            _t = -1

        return {'val': _t, 'unit': 'C'}

    @staticmethod
    def mem_total():
        """
         - total:
           total physical memory available.
        """
        return {'val': psutil.virtual_memory().total, 'unit': 'bytes'}

    @staticmethod
    def mem_available():
        """
         - available:
           the memory that can be given instantly to processes without the
           system going into swap.
           This is calculated by summing different memory values depending
           on the platform and it is supposed to be used to monitor actual
           memory usage in a cross platform fashion.
        """
        return {'val': psutil.virtual_memory().available, 'unit': 'bytes'}

    @staticmethod
    def mem_used_percent():
        """
        - used:
          the percentage usage calculated as (total - available) / total * 100
        """
        return {'val': psutil.virtual_memory().percent, 'unit': '%'}

    @staticmethod
    def mem_used_bytes():
        """
         - used:
            memory used, calculated differently depending on the platform and
            designed for informational purposes only:
            macOS: active + wired
            BSD: active + wired + cached
            Linux: total - free
        """
        return {'val': psutil.virtual_memory().used, 'unit': 'bytes'}

    @staticmethod
    def mem_free():
        """
         - free:
           memory not being used at all (zeroed) that is readily available;
           note that this doesn't reflect the actual memory available
           (use 'available' instead)

        """
        return {'val': psutil.virtual_memory().free, 'unit': 'bytes'}

    @staticmethod
    def platform():
        _info = platform.uname()
        if _info is None or len(_info._fields) == 0:
            return {'val': None, 'unit': None}

        _values = dict()
        for _f in _info._fields:
            _values[_f] = _info.__getattribute__(_f)
        return {'val': _values, 'unit': 'json'}

    @staticmethod
    def network_download_speed() -> dict or None:
        """
            Method to get download speed by testing real network speed.
            It requires speedtest python  system app to be installed.

            WARNING: The method is very slow
        :return:
        """
        try:
            with os.popen(
                StatusInfo.SPEEDTEST_CMD + ' --no-upload --simple '
            ) as speedtest_output:
                for line in speedtest_output:
                    label, value, unit = line.split()
                    if 'download' in label.lower():
                        return {'val': float(value), 'unit': unit}
        except Exception as ex:
            return {'val': float(0), 'unit': 'not installed'}

    @staticmethod
    def network_upload_speed() -> dict:
        """
            Method to get upload speed by testing real network speed.
            It requires speedtest python  system app to be installed.

            WARNING: The method is very slow
        :return:
        """
        try:
            with os.popen(
                StatusInfo.SPEEDTEST_CMD + ' --no-download --simple '
            ) as speedtest_output:
                for line in speedtest_output:
                    label, value, unit = line.split()
                    if 'upload' in label.lower():
                        return {'val': float(value), 'unit': unit}
        except Exception as ex:
            return {'val': float(0), 'unit': 'not installed'}

    @staticmethod
    def all(
        include_network_download_speed: bool = False,
        include_network_upload_speed: bool = False,
    ):

        _info = {
            'memory': {
                'total': StatusInfo.mem_total(),
                'available': StatusInfo.mem_available(),
                'free': StatusInfo.mem_free(),
                'used_bytes': StatusInfo.mem_used_bytes(),
                'used_percents': StatusInfo.mem_used_percent(),
            },
            'disk': {
                'free': StatusInfo.disk_free_bytes(),
                'used': StatusInfo.disk_used_percents(),
            },
            'cpu': {
                'used': StatusInfo.cpu_used_percents(),
                'temp': StatusInfo.cpu_temp(),
            },
            'platform': StatusInfo.platform(),
        }

        if include_network_download_speed or include_network_upload_speed:
            _info['network'] = dict()

        if include_network_download_speed:
            _info['network']['download'] = StatusInfo.network_download_speed()

        if include_network_upload_speed:
            _info['network']['upload'] = StatusInfo.network_upload_speed()

        return _info


_ITEMGETTER_0 = operator.itemgetter(0)


class SyncQueue:
    def __init__(self):
        self.lock = threading.Lock()
        self.queue = deque()

    def len(self):
        with self.lock:
            return len(self.queue)

    def popleft(self):
        with self.lock:
            return self.queue.popleft()

    def popleft_no_block(self):
        with self.lock:
            if len(self.queue):
                return self.queue.popleft()

    def clear(self):
        with self.lock:
            return self.queue.clear()

    def append(self, item):
        with self.lock:
            return self.queue.append(item)


class StatusInfoReader(SourceReader, threading.Thread):
    _index: int = 0
    _data: SyncQueue = SyncQueue()
    _stop: threading.Event = threading.Event()
    _bt: int = None

    def __init__(
        self,
        read_interval_ms=DFLT_STATUS_INFO_READ_INTERVAL,
        include_network_download_speed: bool = False,
        include_network_upload_speed: bool = False,
    ):

        self.read_interval_ms = read_interval_ms
        self.include_network_download_speed = include_network_download_speed
        self.include_network_upload_speed = include_network_upload_speed

        threading.Thread.__init__(self, daemon=True)

    def open(self):
        self._data.clear()
        self._bt = self.get_timestamp()
        self._index = 0

        self._stop.clear()
        self.start()

    def read(self):
        """Returns one data item

        :return: (index, timestamp, character)
        """
        return self._data.popleft_no_block()

    def close(self):
        self._stop.set()

    @property
    def info(self) -> dict:
        return {'bt': self._bt}

    def key(self, data: Any) -> ComparableType:
        """
        :param data: (index, timestamp, character)
        :return: index
        """
        return _ITEMGETTER_0(data)

    def run(self):
        try:
            while not self._stop.is_set():

                self._data.append(
                    (
                        self._index,
                        self.get_timestamp(),
                        StatusInfo.all(
                            self.include_network_download_speed,
                            self.include_network_upload_speed,
                        ),
                    )
                )  # (index, timestamp, character)
                self._index += 1
                if self.read_interval_ms > 0:
                    time.sleep(self.read_interval_ms / 1000)

        except Exception:
            self.close()
            raise


if __name__ == '__main__':

    # As separated info
    # print(f"Memory:\n---------------------------------")
    # print(f"\tmem total {StatusInfo.mem_total()}")
    # print(f"\tmem available {StatusInfo.mem_available()}")
    # print(f"\tmem free {StatusInfo.mem_free()}")
    # print(f"\tmem used bytes {StatusInfo.mem_used_bytes()}")
    # print(f"\tmem used percents {StatusInfo.mem_used_percent()}")
    #
    # print(f"\nDisk:\n---------------------------------")
    # print(f"\tdisk free {StatusInfo.disk_free_bytes()}")
    # print(f"\tdisk used {StatusInfo.disk_used_percents()}")
    #
    # print(f"\nCPU:\n---------------------------------")
    # print(f"\tcpu used {StatusInfo.cpu_used_percents()}")
    # print(f"\tcpu temp {StatusInfo.cpu_temp()}")
    #
    # print(f"\nPlatform:\n---------------------------------")
    # print(f"\tplatform info {StatusInfo.platform()}")
    #
    # print(f"\nNetwork:\n---------------------------------")
    # print(f"\tDownload speed {StatusInfo.network_download_speed()}")
    # print(f"\tUpload speed {StatusInfo.network_upload_speed()}")

    # or as a single JSON
    # print(StatusInfo.all(include_network_download_speed=True, include_network_upload_speed=True))

    with StatusInfoReader() as source:
        print('Ctrl+C to exit')
        time.sleep(3)
        while True:
            try:
                data = source.read()
                if data is not None:
                    index, timestamp, info = data
                    pprint(f'{index}.{timestamp}: {info}')

            except KeyboardInterrupt as kb:
                break

    print('Done!')



__all__ = ['StreamBuffer']

import threading
import time

from stream2py.utility.locked_sorted_deque import RWLockSortedDeque
from stream2py.utility.typing_hints import Optional, Union

from stream2py import *

import logging

logger = logging.getLogger(__name__)

DFLT_SLEEP_TIME_ON_READ_NONE_S = 0.3


class _SourceBuffer:
    def __init__(
        self,
        source_reader_info: dict,
        stop_event: threading.Event,
        *,
        key=None,
        maxlen: int = 10000,
    ):
        """StreamBuffer helper class
        Thread safe buffer for reading and writing data items

        source_name should default to source class name
        source_info should default to args for constructing source class
        an audio source info should have info about sample rates, bit depth, etc.
        reader should get both source_name and source_info
        """
        self._stop_event = stop_event
        self._buffer = RWLockSortedDeque([], key=key, maxlen=maxlen)
        self._source_reader_info = source_reader_info

    def __len__(self):
        return len(self._buffer)

    @property
    def key(self):
        return self._buffer.key

    @property
    def source_reader_info(self):
        return self._source_reader_info

    def append(self, item):
        with self._buffer.writer_lock() as writer:
            writer.append(item)

    def drop(self, n=1):
        with self._buffer.writer_lock() as writer:
            writer.drop(n)

    def mk_reader(self):
        return BufferReader(
            buffer=self._buffer,
            source_reader_info=self._source_reader_info,
            stop_event=self._stop_event,
        )


class StreamBuffer:
    """Handles starting and stopping SourceReader and making BufferReaders

    >>> from stream2py import StreamBuffer
    >>> from stream2py.examples.source_reader import SimpleCounterString
    >>>
    >>> source_reader = SimpleCounterString(start=0, stop=100)
    >>> stream_buffer = StreamBuffer(source_reader=source_reader, maxlen=100)
    >>> stream_buffer.is_running
    False
    >>> print(stream_buffer.source_reader_info)
    None
    >>> stream_buffer.start()
    >>> assert stream_buffer.is_running
    >>> stream_buffer.source_reader_info
    {'start': 0, 'stop': 100, 'open_count': 1}
    >>> open1_reader1 = stream_buffer.mk_reader()
    >>> open1_reader1.next()
    's0'
    >>> open1_reader2 = stream_buffer.mk_reader()
    >>> assert open1_reader1.is_same_buffer(open1_reader2) is True  # readers from the same open instance are the same
    >>> assert open1_reader1.next() != open1_reader2.next()  # item cursor position is different
    >>> stream_buffer.stop()
    >>>
    >>> with stream_buffer:
    ...     stream_buffer.source_reader_info
    ...     open2_reader1  = stream_buffer.mk_reader()
    ...     open2_reader2  = stream_buffer.mk_reader()
    ...     assert open2_reader1.is_same_buffer(open2_reader2) is True  # readers from the same open instance are the same
    ...     assert open2_reader1.is_same_buffer(open1_reader1) is False  # readers from the different open instances
    {'start': 0, 'stop': 100, 'open_count': 2}
    """

    def __init__(
        self,
        source_reader: SourceReader,
        *,
        maxlen: int,
        sleep_time_on_read_none_s: Optional[Union[int, float]] = None,
        auto_drop=True,
    ):
        """
        TODO: option to auto restart source on read exception

        :param source_reader: instance of a SourceReader subclass
        :param maxlen: max number of read data points to store in buffer before data starts dropping off the queue
        :param sleep_time_on_read_none_s: Seconds to sleep when reading None from source_reader. None to use defaults.
        :param auto_drop: False to stop reading when buffer is full and use StreamBuffer.drop() to manually make space.
        """
        assert isinstance(
            source_reader, SourceReader
        ), 'source_reader is not a subclass of SourceReader'
        self.source_reader = source_reader
        self._maxlen = maxlen
        self.auto_drop = auto_drop
        if isinstance(sleep_time_on_read_none_s, (int, float)):
            self._sleep_time_on_read_none_s = sleep_time_on_read_none_s
        elif isinstance(source_reader.sleep_time_on_read_none_s, (int, float)):
            self._sleep_time_on_read_none_s = source_reader.sleep_time_on_read_none_s
        else:
            self._sleep_time_on_read_none_s = DFLT_SLEEP_TIME_ON_READ_NONE_S  # default

        self._stop_event = None
        self.source_buffer = None
        self.start_lock = (
            threading.Lock()
        )  # used to lock mk_reader while source is still starting up
        self._next_reader = None

    def __iter__(self):
        reader = self.mk_reader()
        yield from iter(reader)

    def __next__(self):
        if self._next_reader is None:
            self._next_reader = self.mk_reader()
        return next(self._next_reader)

    def drop(self, n=1):
        """Manually drop items from buffer when auto_drop is False

        :param n: number of items to drop from the left side
        :return:
        """
        if self.auto_drop is True:
            raise RuntimeError(
                'auto_drop must be False to manually drop items from buffer'
            )
        self.source_buffer.drop(n)

    def start(self):
        """Open and start reading from source_reader into buffer"""
        with self.start_lock:
            if self._stop_event and not self._stop_event.is_set():
                self.stop()
            self._set_read_to_buffer_thread_and_source_buffer()
            self._open()
            self._read_to_buffer_thread.start()

    def stop(self):
        """Stop reading and close source_reader"""
        self._stop_event.set()
        time.sleep(1)
        self._next_reader = None

    def mk_reader(self) -> BufferReader:
        """ Makes a BufferReader instance for the currently running  StreamBuffer.
        Reader must be made after start() to have data from said start.

        :return: BufferReader instance
        """
        with self.start_lock:
            if not isinstance(self.source_buffer, _SourceBuffer):
                raise RuntimeError('Readers should be made after starting')
            return self.source_buffer.mk_reader()

    @property
    def source_reader_info(self) -> Optional[dict]:
        """A dict with important source info set by SourceReader.

        :return: dict or None
        """
        if self.source_buffer:
            return self.source_buffer.source_reader_info

    @property
    def is_running(self) -> bool:
        """Checks if stop event has been set.

        :return: bool
        """
        with self.start_lock:
            return self._stop_event is not None and not self._stop_event.is_set()

    @property
    def join(self):
        return self._read_to_buffer_thread.join

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def _run(self):
        try:
            while not self._stop_event.is_set():
                # check if buffer is full and skip read or append
                if self.auto_drop is True or len(self.source_buffer) < self._maxlen:
                    data = self.source_reader.read()
                else:
                    data = None

                if data is not None:
                    self.source_buffer.append(data)
                else:
                    time.sleep(self._sleep_time_on_read_none_s)
        finally:
            if not self._stop_event.is_set():
                self._stop_event.set()
            try:
                self.source_reader.close()
            except Exception as e:
                logger.error(e)

    def _set_read_to_buffer_thread_and_source_buffer(self):
        """Initialize source and buffer
        Makes a new run thread, and stop event that is unique to this run session"""
        if self._stop_event and not self._stop_event.is_set():
            self.stop()
        self.source_buffer = None
        self._read_to_buffer_thread = threading.Thread(target=self._run, daemon=True)
        self._stop_event = threading.Event()

    def _open(self):
        """First thing called after start/run
        Calls source_reader.open() and then sets up source_buffer with latest source_reader.info"""
        self.source_reader.open()
        self.source_buffer = _SourceBuffer(
            source_reader_info=self.source_reader.info,
            stop_event=self._stop_event,
            key=self.source_reader.key,
            maxlen=self._maxlen,
        )


if __name__ == '__main__':
    import operator

    class TenthSecondCounter(SourceReader):
        """Example SourceReader
        Start counting when as soon as you construct"""

        def __init__(self, starting_count=0):
            self._init_kwargs = {
                k: v for k, v in locals().items() if k not in ('self', '__class__')
            }
            self.starting_count = starting_count
            self._count = 0
            self._start_time = 0

        def open(self):
            """Reset params for first read"""
            self._count = self.starting_count
            self._start_time = int(time.time() * 10)

        @property
        def info(self):
            _info = self._init_kwargs.copy()
            _info.update(name=self.__class__.__name__, bt=self._start_time * 100000)
            return _info

        def close(self):
            """Not needed but satisfies the abstract"""

        def read(self):
            next_count = self._count
            next_time = self._start_time + self._count
            now = time.time() * 10
            if now >= next_time:
                self._count += 1
                return next_time * 100000, next_count  # (timestamp_us, count)

        key = operator.itemgetter(0)  # (timestamp_us, count) -> timestamp_us

    sc = TenthSecondCounter()  # source
    sc_buf = StreamBuffer(source_reader=sc, maxlen=100)  # buffer

    # start source, print some data, and check last_item cursor is working
    sc_buf.start()  # start reader to buffer thread
    sc_reader = (
        sc_buf.mk_reader()
    )  # reader must be made after start to have data related to said start
    time.sleep(2)
    assert sc_reader.source_reader_info['bt'] != 0, sc_reader.source_reader_info
    print('data as shown is (timestamp, count) at every tenth of a second')
    while True:
        x = sc_reader.next(ignore_no_item_found=True)
        if x:
            last_read = x
            print(x)
        else:
            break
    assert sc_reader.last_item == last_read, 'last_item did not follow last next() call'

    # check range works and last_item cursor is working
    rstart = 5
    rstop = 10
    source_info = sc_reader.source_reader_info
    start_key = source_info['bt'] + 1e5 * rstart  # 5
    stop_key = source_info['bt'] + 1e5 * rstop  # 10
    range_data = sc_reader.range(start_key, stop_key)
    for expected_data, rdata in zip(range(rstart, rstop + 1), range_data):
        timestamp, data = rdata
        assert data == expected_data, 'wrong data does not match'
    assert (
        expected_data == sc_reader.last_item[1]
    ), f'last_item did not follow last range({start_key}, {stop_key}) call'
    previous_last_item = expected_data

    # check range works with step and peek, and last_item does not move with peek=True
    rstart = 10
    rstop = 15
    rstep = 2
    source_info = sc_reader.source_reader_info
    start_key = source_info['bt'] + 1e5 * rstart  # 5
    stop_key = source_info['bt'] + 1e5 * rstop  # 10
    range_data = sc_reader.range(start_key, stop_key, step=rstep, peek=True)
    for expected_data_value, rdata in zip(range(rstart, rstop + 1, rstep), range_data):
        timestamp, data = rdata
        assert data == expected_data_value, 'wrong data does not match'
    assert (
        previous_last_item == sc_reader.last_item[1]
    ), f'last_item moved but should not when peek=True'

    # stop source and check if reader see it
    sc_reader12 = sc_buf.mk_reader()  # reader1
    assert sc_reader.is_same_buffer(sc_reader12), 'first readers should be equal'
    assert (
        sc_reader.last_item != sc_reader12.last_item
    ), 'first readers should have a different last_item cursor position'
    assert sc_reader.is_stopped is False, 'Reader should see source is not stopped'
    sc_buf.stop()
    assert sc_reader.is_stopped is True, 'Reader should see source is stopped'

    # restart source and make new readers then check old reader is different from new reader
    with sc_buf:  # start and stop with a context manager

        sc_reader21 = sc_buf.mk_reader()  # reader2
        sc_reader22 = sc_buf.mk_reader()  # reader2
        assert sc_reader21.is_same_buffer(
            sc_reader22
        ), 'two new readers should be equal'
        assert not sc_reader.is_same_buffer(
            sc_reader21
        ), 'first reader should not equal new reader'
        assert (
            sc_reader.is_stopped is True
        ), 'first reader should still see a stopped source'
        assert (
            sc_reader21.is_stopped is False
        ), 'new reader should see the source is running'
        assert (
            sc_reader22.last_item == sc_reader21.last_item
        ), 'new readers should now have the same last_item cursor position'

        # print some data from old and new reader and see they are different
        time.sleep(1.5)
        print('comparing first start reader values to second start')
        while True:
            r22 = sc_reader22.next(ignore_no_item_found=True)
            r12 = sc_reader12.next(ignore_no_item_found=True)

            if r22:

                def zero_if_none(val):
                    if val is None:
                        return 0, 0
                    return val

                print(f'{r22} > {r12}, ts_diff={r22[0] - r12[0]}')
                assert r22 > zero_if_none(r12)
            else:
                break
        assert (
            sc_reader22.last_item != sc_reader21.last_item
        ), 'new readers should now have a different last_item cursor position'



"""
PyAudio Source
--------------
This source requires PyAudio.
`Find PyAudio installation instructions here <https://people.csail.mit.edu/hubert/pyaudio/>`_.
::
    pip install PyAudio

.. autoclass:: stream2py.sources.audio.PyAudioSourceReader()
    :members:
    :show-inheritance:

    .. automethod:: __init__

.. autoclass:: stream2py.sources.audio.PaStatusFlags
    :members:
    :undoc-members:
    :show-inheritance:
    :member-order: bysource


PyAudio Mixins
--------------
Example of how to use mixins to manipulate behavior to suit your needs.

.. autoclass:: stream2py.sources.audio.FillErrorWithZeroesMixin
    :members:

.. autoclass:: stream2py.sources.audio.DictDataMixin
    :members:

.. autoclass:: stream2py.sources.audio.RaiseRuntimeErrorOnStatusFlagMixin
    :members:

Example PyAudio Mixin Usage
---------------------------
Mixins must be subclassed in order of precedence from left to right.
Meaning the base class at the right and mixins with overloading methods to the left.

.. autoclass:: stream2py.sources.audio.PyAudioSourceReaderRaiseOnError
    :members:
    :show-inheritance:

.. autoclass:: stream2py.sources.audio.PyAudioSourceReaderRaiseOnError
    :members:
    :show-inheritance:

.. autoclass:: stream2py.sources.audio.PyAudioSourceReaderWithZeroedErrorsAndDictData
    :members:
    :show-inheritance:
"""

__all__ = ['PyAudioSourceReader', 'PaStatusFlags', 'PaCallbackReturnCodes']

from collections import deque
from contextlib import suppress, contextmanager
from enum import IntFlag
import math
import operator
import threading

import pyaudio

from stream2py import SourceReader
from stream2py.utility.typing_hints import Generator, ComparableType, List

_ITEMGETTER_0 = operator.itemgetter(0)


def list_recording_device_index_names():
    """List (index, name) of available recording devices"""
    return sorted(
        (d['index'], d['name'])
        for d in PyAudioSourceReader.list_device_info()
        if d['maxInputChannels'] > 0
    )


def _a_pyaudio_source_reader_can_be_constructed(
    input_device_index,
    rate=44100,
    width=2,
    unsigned=True,
    channels=1,
    frames_per_buffer=1024,
) -> bool:
    try:
        PyAudioSourceReader(**locals())
        return True
    except ValueError:
        return False


# TODO: Merge with find_a_device_index
def find_a_default_input_device_index(verbose=True):
    """Try to find a device (index) that might work for audio input.
    Look for one that has 'microphone' in it's name, if not 'mic', and if not, as a last
    resort, just anything for which a PyAudioSourceReader can be made.
    """
    for index, name in list_recording_device_index_names():
        if 'microphone' in name.lower():
            if verbose:
                print(
                    f'Found {name}. '
                    f"Will use it as the default input device. It's index is {index}"
                )
            return index
    for index, name in list_recording_device_index_names():
        if 'mic' in name.lower():
            if verbose:
                print(
                    f'Found {name}. '
                    f"Will use it as the default input device. It's index is {index}"
                )
            return index
    for index, name in list_recording_device_index_names():
        if _a_pyaudio_source_reader_can_be_constructed(index):
            if verbose:
                print(
                    f'Found {name}. '
                    f"Will use it as the default input device. It's index is {index}"
                )
            return index


# TODO: Test and merge with find_a_default_input_device_index
def find_a_device_index(filt='microphone', dflt=None):
    if isinstance(filt, str):
        match_str = filt

        def filt(x):
            return match_str in x.get('name', match_str).lower()

    match = next(filter(filt, PyAudioSourceReader.list_device_info()), None)
    return (match is not None and match['index']) or dflt


class PaStatusFlags(IntFlag):
    """Enum to check status_flag for error codes

    >>> from stream2py.sources.audio import PaStatusFlags
    >>> PaStatusFlags(0)
    <PaStatusFlags.paNoError: 0>
    >>> PaStatusFlags(2)
    <PaStatusFlags.paInputOverflow: 2>
    >>> PaStatusFlags(3)
    <PaStatusFlags.paInputOverflow|paInputUnderflow: 3>
    >>> PaStatusFlags.paInputOverflow in PaStatusFlags(3)  # Check if contains certain error
    True
    >>> PaStatusFlags.paNoError == PaStatusFlags(3)  # Check for no error
    False
    """

    paNoError = pyaudio.paNoError
    paInputUnderflow = pyaudio.paInputUnderflow
    paInputOverflow = pyaudio.paInputOverflow
    paOutputUnderflow = pyaudio.paOutputUnderflow
    paOutputOverflow = pyaudio.paOutputOverflow
    paPrimingOutput = pyaudio.paPrimingOutput


class PaCallbackReturnCodes(IntFlag):
    """Enum of valid _stream_callback return codes.
    Only used by PyAudioSourceReader._stream_callback"""

    paContinue = pyaudio.paContinue
    paComplete = pyaudio.paComplete
    paAbort = pyaudio.paAbort


class PyAudioSourceReader(SourceReader):
    _pyaudio_instance = None

    def __init__(
        self,
        *,
        rate=44100,
        width=2,
        unsigned=True,
        channels=1,
        input_device_index=None,
        frames_per_buffer=1024,
    ):
        """

        :param rate: Specifies the desired sample rate (in Hz)
        :param width: Sample width in bytes (1, 2, 3, or 4)
        :param unsigned: For 1 byte width, specifies signed or unsigned format. Ignored if byte width is not 1.
        :param channels: The desired number of input channels. Ignored if input_device is not specified (or None).
        :param input_device_index: Index of Input Device to use. Unspecified (or None) uses default device.
        :param frames_per_buffer: Specifies the number of frames per buffer.

        .. todo::
            * option to get input device by name instead of index and warn of duplicate names or missing
            * input_device_index is used by info and must be found and filled into _init_kwargs if looking up by name
            * warn if index does not match name when both name and index are specified
            * option to get input from file
        """
        self._init_kwargs = {
            k: v for k, v in locals().items() if k not in ('self', '__class__')
        }
        with self._pyaudio() as pa:
            input_format = pa.get_format_from_width(width, unsigned)
            pa.is_format_supported(
                rate=rate,
                input_device=input_device_index,
                input_channels=channels,
                input_format=input_format,
                output_device=None,
                output_channels=None,
                output_format=None,
            )

        self._pyaudio_open_params = {
            'rate': rate,
            'channels': channels,
            'format': input_format,
            'input': True,
            'output': False,
            'input_device_index': input_device_index,
            'output_device_index': None,
            'frames_per_buffer': frames_per_buffer,
            'start': True,
            'input_host_api_specific_stream_info': None,
            'output_host_api_specific_stream_info': None,
            'stream_callback': self._stream_callback,
        }

        self._fp = None
        self._first_time_info = None
        self.frame_index = 0
        self.bt = -1
        self._start_time = self.bt
        self.data = None

        self._init_vars()

    def _init_vars(self):
        if self._fp:
            self.close()
        self._fp = None
        self._first_time_info = None
        self.frame_index = 0
        self.bt = -1
        self._start_time = self.bt
        if self.data:
            self.data.clear()
        self.data = deque()

    @property
    def sleep_time_on_read_none_s(self) -> float:
        """One third of the expected rate frames_per_buffer will be filled"""
        seconds_per_read = (
            self._init_kwargs['frames_per_buffer'] / self._init_kwargs['rate']
        )
        one_third_read_rate = seconds_per_read / 3
        return one_third_read_rate

    @property
    def info(self) -> dict:
        """
        Provides a dict with init kwargs, bt, and device info

        >>> from stream2py.sources.audio import PyAudioSourceReader
        >>> from pprint import pprint
        >>> source = PyAudioSourceReader(
        ... rate=44100, width=2, channels=1,
        ... input_device_index=7, frames_per_buffer=4096)  # doctest: +SKIP
        >>> source.open()  # doctest: +SKIP
        >>> pprint(source.info) # doctest: +SKIP
        {'bt': 1582851038965183,
         'channels': 1,
         'device_info': {'defaultHighInputLatency': 0.021333333333333333,
                         'defaultHighOutputLatency': 0.021333333333333333,
                         'defaultLowInputLatency': 0.021333333333333333,
                         'defaultLowOutputLatency': 0.021333333333333333,
                         'defaultSampleRate': 48000.0,
                         'hostApi': 0,
                         'index': 7,
                         'maxInputChannels': 128,
                         'maxOutputChannels': 128,
                         'name': 'sysdefault',
                         'structVersion': 2},
         'frames_per_buffer': 4096,
         'input_device_index': 7,
         'rate': 44100,
         'unsigned': True,
         'width': 2}
        >>> source.close()  # doctest: +SKIP

        :return: dict
        """
        _info = {'bt': self.bt}
        _info.update(**self._init_kwargs)
        with suppress(Exception):
            _info.update(
                device_info=self._pyaudio_instance.get_device_info_by_index(
                    self._init_kwargs['input_device_index']
                )
            )
        return _info

    def key(self, data) -> ComparableType:
        """
        :param data: (timestamp, waveform, frame_count, time_info, status_flags)
        :return: timestamp
        """
        return _ITEMGETTER_0(data)

    def data_to_append(self, timestamp, waveform, frame_count, time_info, status_flags):
        """Can to be overloaded to change the shape of read outputs by altering the return value.
        The key function must also be overloaded to return timestamp from the new shape.

        :param timestamp: start time of waveform
        :param waveform: recorded input data
        :param frame_count: number of frames, sample count
        :param time_info: dict, see http://portaudio.com/docs/v19-doxydocs/structPaStreamCallbackTimeInfo.html
        :param status_flags: PaStatusFlags
        :return: (timestamp, waveform, frame_count, time_info, status_flags)
        """
        return timestamp, waveform, frame_count, time_info, status_flags

    def open(self):
        """Will first close if already open and clear any data before starting the audio stream"""
        self._init_vars()
        self._init_pyaudio()
        self.bt = self.get_timestamp()
        self._start_time = self.bt
        self._fp = self._pyaudio_instance.open(**self._pyaudio_open_params)

    def close(self):
        """Stop audio stream"""
        with suppress(Exception):
            self._fp.stop_stream()
        with suppress(Exception):
            self._fp.close()
        self._terminate_pyaudio()

    def read(self):
        """Returns one data item as structured by PyAudioSourceReader.data_to_append

        :return: (timestamp, waveform, frame_count, time_info, status_flags)
        """
        if len(self.data):
            return self.data.popleft()

    def _stream_callback(self, in_data, frame_count, time_info, status_flags):
        """Calculates timestamp based on open() bt and frames read.
        If there is an error conveyed by status_flags, the frame count is reset to 0 and starting timestamp is shifted
        from open() bt by time_info to approximate actual time in case of sample loss.
        See _stream_callback in https://people.csail.mit.edu/hubert/pyaudio/docs/#class-stream

        :param in_data: recorded input data, waveform
        :param frame_count: number of frames, sample count
        :param time_info: dictionary
        :param status_flags: PaStatusFlags
        :return: None, PaCallbackReturnCodes.paContinue
        """
        if not self._first_time_info:
            self._first_time_info = time_info

        if self.frame_index == 0:
            # set start time based on audio time_info difference
            _time_info_diff_s = (
                time_info['input_buffer_adc_time']
                - self._first_time_info['input_buffer_adc_time']
            )
            _timestamp_diff = (
                _time_info_diff_s * self.timestamp_seconds_to_unit_conversion
            )
            self._start_time = self.bt + _timestamp_diff

        _frame_time_s = self.frame_index / self._init_kwargs['rate']
        timestamp = int(
            self._start_time + _frame_time_s * self.timestamp_seconds_to_unit_conversion
        )

        self.data.append(
            self.data_to_append(
                timestamp, in_data, frame_count, time_info, status_flags
            )
        )

        if PaStatusFlags(status_flags) != PaStatusFlags.paNoError:
            # reset frame index and thus self._start_time on any error status
            self.frame_index = 0
        else:
            self.frame_index += frame_count
        return None, PaCallbackReturnCodes.paContinue

    @classmethod
    def _init_pyaudio(cls) -> pyaudio.PyAudio:
        cls._terminate_pyaudio()
        cls._pyaudio_instance = pyaudio.PyAudio()
        return cls._pyaudio_instance

    @classmethod
    def _terminate_pyaudio(cls):
        with suppress(Exception):
            if cls._pyaudio_instance is not None and isinstance(
                cls._pyaudio_instance, (pyaudio.PyAudio,)
            ):
                cls._pyaudio_instance.terminate()
                cls._pyaudio_instance = None

    @classmethod
    @contextmanager
    def _pyaudio(cls) -> Generator[pyaudio.PyAudio, None, None]:
        try:
            yield cls._init_pyaudio()
        finally:
            cls._terminate_pyaudio()

    @classmethod
    def list_device_info(cls) -> List[dict]:
        """
        .. todo::
            * filter for only devices with input channels

        :return: list
        """
        with cls._pyaudio() as pa:
            return [
                pa.get_device_info_by_index(idx) for idx in range(pa.get_device_count())
            ]

    @staticmethod
    def audio_buffer_size_seconds_to_maxlen(
        buffer_size_seconds, rate, frames_per_buffer
    ) -> int:
        """Calculate maxlen for StreamBuffer to keep a minimum of buffer_size_seconds of data on buffer

        :param buffer_size_seconds: desired length of StreamBuffer in seconds
        :param rate: sample rate
        :param frames_per_buffer: number of frames per buffer
        :return: maxlen for StreamBuffer
        """
        seconds_per_read = frames_per_buffer / rate
        return math.ceil(buffer_size_seconds / seconds_per_read)

    @classmethod
    def list_recording_devices(cls):
        """List names of available recording devices"""
        return sorted(
            d['name'] for d in cls.list_device_info() if d['maxInputChannels'] > 0
        )


class FillErrorWithZeroesMixin:
    """Mixin to handle all pyaudio's error status flags by filling the error time gap with zeros"""

    _first_error_timestamp = None
    _error_status_flag = None

    def _stream_callback(self, in_data, frame_count, time_info, status_flags):
        """On status flag error code, reate wf bytes of value zero for the entire duration of the error to replace
        garbled data in addtion to the base behavior.

        :param in_data: recorded input data, waveform
        :param frame_count: number of frames, sample count
        :param time_info: dictionary
        :param status_flags: PaStatusFlags
        :return: None, PaCallbackReturnCodes.paContinue
        """
        if not self._first_time_info:
            self._first_time_info = time_info

        if self.frame_index == 0:
            # set start time based on audio time_info difference
            _time_info_diff_s = (
                time_info['input_buffer_adc_time']
                - self._first_time_info['input_buffer_adc_time']
            )
            _timestamp_diff = (
                _time_info_diff_s * self.timestamp_seconds_to_unit_conversion
            )
            self._start_time = self.bt + _timestamp_diff

        _frame_time_s = self.frame_index / self._init_kwargs['rate']
        timestamp = int(
            self._start_time + _frame_time_s * self.timestamp_seconds_to_unit_conversion
        )

        if PaStatusFlags(status_flags) != PaStatusFlags.paNoError:
            # reset frame index and thus self._start_time on any error status
            self.frame_index = 0
            if self._first_error_timestamp is None:
                self._first_error_timestamp = timestamp  # track when errors started
                self._error_status_flag = PaStatusFlags(
                    status_flags
                )  # track what errors occured
            else:
                self._error_status_flag |= PaStatusFlags(
                    status_flags
                )  # use OR to mark any new error status flags
        else:
            if (
                self._first_error_timestamp is not None
            ):  # first ok status after there was an error status
                (zeroed_data, zeroed_count,) = self._fill_time_interval_with_zeroes(
                    self._first_error_timestamp, timestamp
                )
                self.data.append(
                    self.data_to_append(
                        self._first_error_timestamp,
                        zeroed_data,
                        zeroed_count,
                        time_info,
                        self._error_status_flag,
                    )
                )
                self._first_error_timestamp = None
                self._error_status_flag = None

            self.data.append(
                self.data_to_append(
                    timestamp, in_data, frame_count, time_info, status_flags
                )
            )
            self.frame_index += frame_count
        return None, PaCallbackReturnCodes.paContinue

    def _fill_time_interval_with_zeroes(
        self, first_error_status_ts, first_ok_status_ts
    ):
        """Create wf bytes of value zero for the entire duration of the error to replace garbled data

        :param first_error_status_ts: Error bt
        :param first_ok_status_ts: Error tt
        :return: in_data, frame_count
        """

        single_zero_sample = (
            b'\x00' * self._init_kwargs['channels'] * self._init_kwargs['width']
        )  # interleaved zeros
        samples_per_time_unit = (
            self._init_kwargs['rate'] / self.timestamp_seconds_to_unit_conversion
        )

        delta_time = first_ok_status_ts - first_error_status_ts
        sample_count = int(delta_time * samples_per_time_unit)
        wf_bytes = single_zero_sample * sample_count
        return wf_bytes, sample_count


class FillErrorWithOnesMixin:
    """Mixin to handle all pyaudio's error status flags by filling the error time gap with zeros"""

    _first_error_timestamp = None
    _error_status_flag = None

    def _stream_callback(self, in_data, frame_count, time_info, status_flags):
        """On status flag error code, reate wf bytes of value zero for the entire duration of the error to replace
        garbled data in addtion to the base behavior.

        :param in_data: recorded input data, waveform
        :param frame_count: number of frames, sample count
        :param time_info: dictionary
        :param status_flags: PaStatusFlags
        :return: None, PaCallbackReturnCodes.paContinue
        """
        if not self._first_time_info:
            self._first_time_info = time_info

        if self.frame_index == 0:
            # set start time based on audio time_info difference
            _time_info_diff_s = (
                time_info['input_buffer_adc_time']
                - self._first_time_info['input_buffer_adc_time']
            )
            _timestamp_diff = (
                _time_info_diff_s * self.timestamp_seconds_to_unit_conversion
            )
            self._start_time = self.bt + _timestamp_diff

        _frame_time_s = self.frame_index / self._init_kwargs['rate']
        timestamp = int(
            self._start_time + _frame_time_s * self.timestamp_seconds_to_unit_conversion
        )

        if PaStatusFlags(status_flags) != PaStatusFlags.paNoError:
            # reset frame index and thus self._start_time on any error status
            self.frame_index = 0
            if self._first_error_timestamp is None:
                self._first_error_timestamp = timestamp  # track when errors started
                self._error_status_flag = PaStatusFlags(
                    status_flags
                )  # track what errors occured
            else:
                self._error_status_flag |= PaStatusFlags(
                    status_flags
                )  # use OR to mark any new error status flags
        else:
            if (
                self._first_error_timestamp is not None
            ):  # first ok status after there was an error status
                zeroed_data, zeroed_count = self._fill_time_interval_with_ones(
                    self._first_error_timestamp, timestamp
                )
                self.data.append(
                    self.data_to_append(
                        self._first_error_timestamp,
                        zeroed_data,
                        zeroed_count,
                        time_info,
                        self._error_status_flag,
                    )
                )
                self._first_error_timestamp = None
                self._error_status_flag = None

            self.data.append(
                self.data_to_append(
                    timestamp, in_data, frame_count, time_info, status_flags
                )
            )
            self.frame_index += frame_count
        return None, PaCallbackReturnCodes.paContinue

    def _fill_time_interval_with_ones(
        self, first_error_status_ts, first_ok_status_ts, fill_value=1
    ):
        """Create wf bytes of value zero for the entire duration of the error to replace garbled data

        :param first_error_status_ts: Error bt
        :param first_ok_status_ts: Error tt
        :param fill_value: default is 1
        :return: in_data, frame_count
        """

        single_fill_sample = (
            fill_value.to_bytes(self._init_kwargs['width'], 'little')
            * self._init_kwargs['channels']
        )
        samples_per_time_unit = (
            self._init_kwargs['rate'] / self.timestamp_seconds_to_unit_conversion
        )

        delta_time = first_ok_status_ts - first_error_status_ts
        sample_count = int(delta_time * samples_per_time_unit)
        wf_bytes = single_fill_sample * sample_count
        return wf_bytes, sample_count


class DictDataMixin:
    """Mixin to reduce data to a dict with bt, wf, and status_flag. Removing typically discarded information."""

    def data_to_append(self, timestamp, waveform, frame_count, time_info, status_flags):
        """Simplify data only

        :param timestamp: start time of waveform
        :param waveform: recorded input data
        :param frame_count: discarded data
        :param time_info: discarded data
        :param status_flags: PaStatusFlags error codes
        :return: {'bt': timestamp, 'wf': waveform, 'status_flags': status_flags}
        """
        return {'bt': timestamp, 'wf': waveform, 'status_flags': status_flags}

    def key(self, data) -> ComparableType:
        """
        :param data: {'bt': timestamp, 'wf': waveform, 'status_flags': status_flags}
        :return: data['bt']
        """
        return data['bt']


class RaiseRuntimeErrorOnStatusFlagMixin:
    """Mixin to raise RuntimeError when status_flag is not PaStatusFlags.paNoError (0)"""

    def data_to_append(self, timestamp, waveform, frame_count, time_info, status_flags):
        if PaStatusFlags.paNoError != PaStatusFlags(status_flags):
            raise RuntimeError(PaStatusFlags(status_flags))

        return super(RaiseRuntimeErrorOnStatusFlagMixin, self).data_to_append(
            timestamp, waveform, frame_count, time_info, status_flags
        )


class PyAudioSourceReaderRaiseOnError(
    RaiseRuntimeErrorOnStatusFlagMixin, PyAudioSourceReader
):
    """PyAudioSourceReader changed to handle errors by raising RuntimeError"""


class PyAudioSourceReaderWithZeroedErrorsAndDictData(
    DictDataMixin, FillErrorWithZeroesMixin, PyAudioSourceReader
):
    """PyAudioSourceReader changed to handle errors and serve data in an easy to read dict."""


class PyAudioSourceReaderWithOnedErrorsAndDictData(
    DictDataMixin, FillErrorWithOnesMixin, PyAudioSourceReader
):
    """PyAudioSourceReader changed to handle errors and serve data in an easy to read dict."""


def _test_run_PyAudioSourceReader(
    readerClass=PyAudioSourceReader,
    timestamp_getter=lambda data: data[0],
    in_data_getter=lambda data: data[1],
    status_flags_getter=lambda data: data[4],
):
    """Run PyAudioSourceReader and print timestamp, byte length, status code, status flag

    :param readerClass: PyAudioSourceReader class or subclass
    :param timestamp_getter: get timestamp from data
    :param in_data_getter: get wf bytes from data
    :param status_flags_getter: get status_flags from data
    """
    from pprint import pprint

    source = readerClass(
        rate=44100, width=2, channels=1, input_device_index=0, frames_per_buffer=4096,
    )
    pprint(source.info)
    try:
        source.open()
        pprint(source.info)
        i = 0
        while i < 600:
            data = source.read()
            if data:
                i += 1
                timestamp = timestamp_getter(data)
                in_data = in_data_getter(data)
                status_flags = status_flags_getter(data)
                print(
                    timestamp,
                    len(in_data),
                    int(status_flags),
                    PaStatusFlags(status_flags),
                )

    finally:
        source.close()
        pprint(source.info)


if __name__ == '__main__':
    # _test_run_PyAudioSourceReader()
    # _test_run_PyAudioSourceReader(PyAudioSourceReaderRaiseOnError)
    _test_run_PyAudioSourceReader(
        PyAudioSourceReaderWithZeroedErrorsAndDictData,
        timestamp_getter=lambda data: data['bt'],
        in_data_getter=lambda data: data['wf'],
        status_flags_getter=lambda data: data['status_flags'],
    )



from stream2py import SourceReader


class SimpleCounterString(SourceReader):
    """Count in range returning a string formatted as f's{count}'"""

    def __init__(self, start, stop):
        assert start < stop
        self.start = start
        self.stop = stop
        self.range_iterator = None
        self.open_count = 0

    def open(self):
        self.open_count += 1
        self.range_iterator = iter(range(self.start, self.stop))

    def read(self):
        value = next(self.range_iterator, None)
        if value is not None:
            return f's{value}'
        else:
            return None

    def close(self):
        del self.range_iterator
        self.range_iterator = None

    @property
    def info(self):
        return dict(start=self.start, stop=self.stop, open_count=self.open_count)

    def key(self, data):
        return int(data[1:])



"""
=================================
Triggered Starts in a Typing Test
=================================

.. graphviz::

    digraph components {
        subgraph cluster_0 {
            style=filled;
            color=lightgrey;
            node [style=filled,color=white];
            label="Sources";
            labeljust="l"
            "TypingTest"
            "KeyboardInput"
        }

        subgraph cluster_1 {
            node [style=filled];
            label="Readers";
            labeljust="l"
            color=lightgrey
            node [style=filled,color=lightgrey];
            "TestProctor"
            "PromptSwitch"
        }
        "KeyboardInput" -> "PromptSwitch" -> "TypingTest" -> "TestProctor";
        "KeyboardInput" -> "TestProctor";
    }

This example demonstrates the usage of stream2py with 4 asynchronous components:

1. KeyboardInputSourceReader starts at the launch of this program and listens to key presses.
2. TypingTest waits for start, stop, and add prompt commands.
3. PromptSwitch reads keyboard inputs for commands and tells TypingTest what to do.
4. TestProctor gives you TypingTest prompts and grades your keyboard input submissions.

Usage
-----
Launching the script:
::
    python -m stream2py.examples.usage.triggered_starts

.. graphviz::

    digraph flow_chart {
        label="User Flow";
        labeljust="l"
        labelloc="t"
        splines=polyline
        nodesep=.2
        {
            node [shape=oval]
            enter [label="Begin"]
            exit [label="Exit"]
        }
        {
            node [shape=diamond label="Is Started?"]
            started_start
            started_stop
            started_other
        }
        {
            node [shape=box]
            start [label="Start"]
            stop [label="Stop"]
            submit [label="Submit for Score"]
            prompt [label="New Test Prompt"]
            usage [label="Display Commands"]
        }
        input [shape=parallelogram, label="Keyboard Input"]
        enter -> usage -> input;
        input -> started_start [ headlabel=<<font color="grey25">"start"</font>>, labeldistance=2];
        input -> started_stop [ headlabel=<<font color="grey25">"stop"</font>>, labeldistance=3 ];
        input -> started_other [ headlabel=<<font color="grey25">"other"*</font>>, labeldistance=2.5 ];
        input -> exit [ headlabel=<<font color="grey25">"exit"</font>>, labeldistance=1.5 ];
        started_start -> submit  [ label="Yes" ];
        started_start -> start [ label="No" ];
        started_stop -> stop [ label="Yes" ];
        started_stop -> usage [ label="No" ];
        started_other -> submit [ label="Yes" ];
        started_other -> usage [ label="No" ];
        submit -> prompt -> input;
        start -> prompt
        stop -> input
    }

Classes
-------

.. autoclass:: stream2py.sources.keyboard_input.KeyboardInputSourceReader
.. autoclass:: TypingTest
.. autoclass:: PromptSwitch
.. autoclass:: TestProctor


.. todo::
    * Generalize event based triggers.
    * Refactor to simplify design patterns.
    * Clarify what components are doing and the objectives of the example.
"""


from collections import deque
from contextlib import contextmanager, suppress
import difflib
import operator
import threading
import time

from stream2py import SourceReader, StreamBuffer, BufferReader
from stream2py.sources.keyboard_input import KeyboardInputSourceReader
from stream2py.utility.typing_hints import (
    ComparableType,
    Any,
    Optional,
    Iterable,
    Union,
)

_ITEMGETTER_0 = operator.itemgetter(0)
_TYPING_TEST_PROMPTS = '''Welcome to the typing test tutorial!
This example demonstrates the usage of stream2py with 4 asynchronous components.
First is KeyboardInputSourceReader which starts at the launch of this program.
Second is TypingTest, also a SourceReader, which queues up test prompts when started.
Third is PromptSwitch, a BufferReader consumer that reads keyboard inputs for commands and tells TypingTest what to do.
Last is TestProctor, a BufferReader consumer that
'''.splitlines(
    keepends=False
)


class TypingTest(
    SourceReader
):  # TODO: make an abc that adds timestamp or counter automatically
    PROMPT_END = None

    def __init__(self, prompts: Iterable[str]):
        self.prompts = prompts
        self.bt = -1

        self.prompt_deque = deque(prompts)
        self.prompt_deque.append(TypingTest.PROMPT_END)

        self._release_lock = threading.Lock()
        self._release_deque = deque()

    def open(self) -> None:
        self._reset_deque_order()
        self.bt = self.get_timestamp()
        with self._release_lock:
            self._release_deque.clear()

    def read(self) -> Optional[Any]:
        if len(self._release_deque) > 0:
            with self._release_lock:
                return self._release_deque.popleft()

    def close(self) -> None:
        pass

    @property
    def info(self) -> dict:
        return {'bt': self.bt, 'prompts': self.prompts}

    def key(self, data: Any) -> ComparableType:
        return _ITEMGETTER_0(data)

    def add_prompt(self):
        with self._release_lock:
            prompt = self.prompt_deque[0]
            timestamp = self.get_timestamp()
            self._release_deque.append((timestamp, prompt))
        self.prompt_deque.rotate(-1)
        return prompt

    def _reset_deque_order(self):
        while self.prompt_deque[-1] is not TypingTest.PROMPT_END:
            self.prompt_deque.rotate(1)


class PromptSwitch(threading.Thread):
    """PromptSwitch will recognize 3 commands: "start", "stop", and "exit" followed by a return key press.

    1. "start" will start adding prompts after each return key press.
    2. "stop" will stop adding prompts
    3. "exit" will exit the program"""

    def __init__(self, input: StreamBuffer, prompts: StreamBuffer):
        if not isinstance(input.source_reader, KeyboardInputSourceReader):
            raise TypeError(
                f'input must be a StreamBuffer with KeyboardInputSourceReader: {input}'
            )
        if not isinstance(prompts.source_reader, TypingTest):
            raise TypeError(
                f'prompts must be a StreamBuffer with TypingTest: {prompts}'
            )
        threading.Thread.__init__(self, daemon=True)
        self.stop_event = threading.Event()

        self.input = input
        self.prompts = prompts

        self.is_started = False

    def stop(self):
        self.stop_event.set()

    def run(self):
        self.display_usage()
        input_reader = self.input.mk_reader()
        for index, timestamp, char in iter(input_reader):
            if self.stop_event.is_set():
                break
            print(
                char, sep='', end='\n' if char == '\r' else '', flush=True
            )  # print input character

            if char == '\r':  # "return" key
                if self._check_trigger_string('exit', input_reader, index):
                    self.stop()
                    break
                elif self.is_started is False:
                    if self._check_trigger_string('start', input_reader, index):
                        self.prompts.start()
                        self.prompts.source_reader.add_prompt()
                        self.is_started = True
                    else:
                        self.display_usage()
                elif self._check_trigger_string('stop', input_reader, index):
                    self.prompts.stop()
                    self.is_started = False
                else:
                    prompt = self.prompts.source_reader.add_prompt()
                    if prompt is TypingTest.PROMPT_END:
                        self.prompts.stop()
                        self.is_started = False
        print('Thanks for playing!')

    @staticmethod
    def display_usage():
        print(
            'Commands that can be entered anytime:\n\r'
            ' start - begin typing test\n\r'
            ' stop  - end typing test\n\r'
            ' exit  - close program\n\r'
        )

    @staticmethod
    def _check_trigger_string(
        trigger_string: str, input_reader: BufferReader, return_index: int
    ):
        start_index = return_index - len(trigger_string)
        stop_index = return_index - 1
        input_data = input_reader.range(start=start_index, stop=stop_index, peek=True)
        input_str = ''.join(c for i, t, c in input_data)
        return input_str == trigger_string


class TestProctor(threading.Thread):
    """Give prompts and grade input"""

    def __init__(self, input: StreamBuffer, prompts: StreamBuffer):
        if not isinstance(input.source_reader, KeyboardInputSourceReader):
            raise TypeError(
                f'input must be a StreamBuffer with KeyboardInputSourceReader: {input}'
            )
        if not isinstance(prompts.source_reader, TypingTest):
            raise TypeError(
                f'prompts must be a StreamBuffer with TypingTest: {prompts}'
            )
        threading.Thread.__init__(self, daemon=True)
        self.stop_event = threading.Event()

        self.input = input
        self.prompts = prompts

    def stop(self):
        self.stop_event.set()

    def run(self):
        while not self.stop_event.is_set():
            if not self.prompts.is_running:
                time.sleep(0.1)
            else:
                prev_timestamp = None
                prev_prompt = None
                for timestamp, prompt in iter(self.prompts):
                    if prev_timestamp:
                        self.grade(prev_prompt, prev_timestamp, timestamp)
                    print('Prompt: ', prompt, sep='', end='\n\r Input: ')
                    prev_timestamp = timestamp
                    prev_prompt = prompt
                else:
                    if prev_timestamp and prev_prompt is not TypingTest.PROMPT_END:
                        self.grade(prev_prompt, prev_timestamp, float('inf'))

    def grade(
        self,
        prompt: str,
        begin_time: Union[int, float],
        end_time: Union[int, float],
        debug=False,
    ):
        input_reader = self.input.mk_reader()
        all_input_data = input_reader.range(0, float('inf'))
        relevant_characters = [
            c for i, t, c in all_input_data if begin_time <= t <= end_time
        ]
        try:
            stop_index = relevant_characters.index('\r')
        except ValueError:
            stop_index = len(relevant_characters)
        relevant_timestamps = [
            t for i, t, c in all_input_data if begin_time <= t <= end_time
        ]
        true_start_time = relevant_timestamps[0]
        true_end_time = relevant_timestamps[stop_index]
        input_string = ''.join(relevant_characters[0:stop_index])
        score = len(prompt) - sum(
            1
            for diff in difflib.ndiff(prompt, input_string)
            if not diff.startswith(' ')
        )
        time_seconds = (true_end_time - true_start_time) / 1e6
        if debug:
            print(
                f'\n\rPrompt: {prompt}\n\r'
                f' Input: {input_string}\n\r'
                f'''Characters: [{', '.join(f'"{ch}"' for ch in input_string)}]\n\r''',
                flush=True,
            )
        print(
            f'You scored {score} out of {len(prompt)} in {round(time_seconds, 1)} seconds!\n\r'
        )


@contextmanager
def source_runner(sources: dict):
    """contextmanager for putting SourceReaders in StreamBuffers and calling start and stop"""
    stream_buffers = {}
    try:
        for name, s in sources.items():
            if not isinstance(s['source'], SourceReader):
                raise TypeError(f'Source must be a SourceReader: {s}')
            stream_buffers[name] = StreamBuffer(
                source_reader=s['source'], maxlen=s['maxlen']
            )
            if s['start'] is True:
                stream_buffers[name].start()
        yield stream_buffers
    finally:
        for name, s_buf in stream_buffers.items():
            with suppress(Exception):
                s_buf.stop()


@contextmanager
def consumer_runner(consumers: dict):
    """contextmanager for calling start and stop on consumers"""
    try:
        for name, c in consumers.items():
            c.start()
        yield consumers
    finally:
        for name, c in consumers.items():
            with suppress(Exception):
                c.stop()


def main():
    sources = {
        'input': dict(source=KeyboardInputSourceReader(), start=True, maxlen=None),
        'prompts': dict(
            source=TypingTest(prompts=_TYPING_TEST_PROMPTS), start=False, maxlen=None,
        ),
    }
    with source_runner(sources) as stream_buffers:
        consumers = {
            'switch': PromptSwitch(
                input=stream_buffers['input'], prompts=stream_buffers['prompts'],
            ),
            'grader': TestProctor(
                input=stream_buffers['input'], prompts=stream_buffers['prompts'],
            ),
        }
        with consumer_runner(consumers):
            consumers['switch'].join()  # wait for PromptSwitch to exit


if __name__ == '__main__':
    main()



from bisect import bisect_left, bisect_right
from collections import deque
from itertools import islice

from stream2py.utility.sorted_collection import SortedCollection


class SortedDeque(SortedCollection):
    """Deque sorted by a key function.

    Intent of SortedDeque is to efficiently search buffered data while efficiently maintaining a fixed length buffer
    that is frequently popping the head and appending to the tail.

    Overloaded SortedCollection methods __contains__, index, and count due to the lack of
    slicing in deques. Efficiency of said methods is still better than base deque.

    Added an append method for cases where the new item to insert is known to be the last.
    """

    def __init__(self, iterable=(), key=None, maxlen=None):
        SortedCollection.__init__(self, iterable, key)
        self._keys = deque(self._keys, maxlen=maxlen)
        self._items = deque(self._items, maxlen=maxlen)
        self._maxlen = maxlen

    def to_jdict(self):
        """TODO: WIP. Need to convert key function and iterable items to json friendly format"""
        return {
            'iterable': list(self._items),
            'key': self._key,
            'maxlen': self._maxlen,
        }

    @classmethod
    def from_jdict(cls, params_dict):
        return cls(**params_dict)

    def _getmaxlen(self):
        return self._maxlen

    def _setmaxlen(self, maxlen):
        if maxlen is not self._maxlen:
            self.__init__(self._items, key=self._key, maxlen=maxlen)

    def _delmaxlen(self):
        self._setmaxlen(None)

    maxlen = property(_getmaxlen, _setmaxlen, _delmaxlen, 'maxlen function')

    def _setkey(self, key):
        if key is not self._key:
            self.__init__(self._items, key=key, maxlen=self._maxlen)

    def clear(self):
        self.__init__([], self._key, self._maxlen)

    def copy(self):
        return self.__class__(self, self._key, self._maxlen)

    def __repr__(self):
        return '%s(%r, key=%s, maxlen=%s)' % (
            self.__class__.__name__,
            list(self._items),  # Not sure if the is the best way to do this
            getattr(self._given_key, '__name__', repr(self._given_key)),
            getattr(self._maxlen, '__name__', repr(self._maxlen)),
        )

    def __reduce__(self):
        return self.__class__, (self._items, self._given_key, self._maxlen)

    def __contains__(self, item):
        k = self._key(item)
        i = bisect_left(self._keys, k)
        j = bisect_right(self._keys, k)
        return item in islice(self._items, i, j)

    def index(self, item):
        """Find the position of an item.  Raise ValueError if not found."""
        k = self._key(item)
        i = bisect_left(self._keys, k)
        j = bisect_right(self._keys, k)
        return self._items.index(item, i, j)

    def count(self, item):
        """Return number of occurrences of item"""
        k = self._key(item)
        i = bisect_left(self._keys, k)
        j = bisect_right(self._keys, k)
        return sum(1 for chk_item in islice(self._items, i, j) if chk_item == item)

    def append(self, item):
        """Append item to the end.  Raise ValueError if item key is less than last item key"""
        k = self._key(item)
        try:
            if k <= self._keys[-1]:
                raise ValueError(
                    'Item key must be greater than last item key to append: %r' % (k,)
                )
        except IndexError as e:
            if len(self._keys) != 0:
                raise e
        self._keys.append(k)
        self._items.append(item)

    def range(self, start, stop, step=None):
        """Return list of items within start and stop key range.

        :param start: starting key
        :param stop: stopping key
        :param step:
        :return: list with items in range
        """
        i = bisect_left(self._keys, start)
        j = bisect_right(self._keys, stop)
        return self.range_by_index(i, j, step)

    def range_by_index(self, start_index, stop_index, step=None):
        """Return list of items within start and stop index range.

        :param start_index: starting index
        :param stop_index: stopping index
        :param step:
        :return: list with items in range
        """
        return list(islice(self._items, start_index, stop_index, step))

    def find_last_gt(self, k):
        """Return last item with a key > k.  Raise ValueError if not found.
        Basically a tail function with a greater than check"""
        try:
            if self._keys[-1] > k:
                return self._items[-1]
        except IndexError as e:
            if len(self._keys) != 0:
                raise e
        raise ValueError('No item found with key above: %r' % (k,))

    def drop(self, n=1):
        """Remove n items from the left

        :param n:
        :return:
        """
        for _ in range(n):
            del self._keys[0]
            del self._items[0]


# ---------------------------  Simple demo and tests  -------------------------
if __name__ == '__main__':
    # ---------------------------  Below tests are copied from SortedCollection  -------------------------

    def ve2no(f, *args):
        """Convert ValueError result to -1"""
        try:
            return f(*args)
        except ValueError:
            return -1

    def slow_index(seq, k):
        """Location of match or -1 if not found"""
        for i, item in enumerate(seq):
            if item == k:
                return i
        return -1

    def slow_find(seq, k):
        """First item with a key equal to k. -1 if not found"""
        for item in seq:
            if item == k:
                return item
        return -1

    def slow_find_le(seq, k):
        """Last item with a key less-than or equal to k."""
        for item in reversed(seq):
            if item <= k:
                return item
        return -1

    def slow_find_lt(seq, k):
        """Last item with a key less-than k."""
        for item in reversed(seq):
            if item < k:
                return item
        return -1

    def slow_find_ge(seq, k):
        """First item with a key-value greater-than or equal to k."""
        for item in seq:
            if item >= k:
                return item
        return -1

    def slow_find_gt(seq, k):
        """First item with a key-value greater-than or equal to k."""
        for item in seq:
            if item > k:
                return item
        return -1

    from random import choice

    pool = [1.5, 2, 2.0, 3, 3.0, 3.5, 4, 4.0, 4.5]
    for i in range(500):
        for n in range(6):
            s = [choice(pool) for i in range(n)]
            sc = SortedDeque(s)
            s.sort()
            for probe in pool:
                assert repr(ve2no(sc.index, probe)) == repr(slow_index(s, probe))
                assert repr(ve2no(sc.find, probe)) == repr(slow_find(s, probe))
                assert repr(ve2no(sc.find_le, probe)) == repr(slow_find_le(s, probe))
                assert repr(ve2no(sc.find_lt, probe)) == repr(slow_find_lt(s, probe))
                assert repr(ve2no(sc.find_ge, probe)) == repr(slow_find_ge(s, probe))
                assert repr(ve2no(sc.find_gt, probe)) == repr(slow_find_gt(s, probe))
            for i, item in enumerate(s):
                assert repr(item) == repr(sc[i])  # test __getitem__
                assert item in sc  # test __contains__ and __iter__
                assert s.count(item) == sc.count(item)  # test count()
            assert len(sc) == n  # test __len__
            assert list(map(repr, reversed(sc))) == list(
                map(repr, reversed(s))
            )  # test __reversed__
            assert list(sc.copy()) == list(sc)  # test copy()
            sc.clear()  # test clear()
            assert len(sc) == 0

    sd = SortedCollection('The quick Brown Fox jumped'.split(), key=str.lower)
    assert sd._keys == ['brown', 'fox', 'jumped', 'quick', 'the']
    assert sd._items == ['Brown', 'Fox', 'jumped', 'quick', 'The']
    assert sd._key == str.lower
    assert (
        repr(sd)
        == "SortedCollection(['Brown', 'Fox', 'jumped', 'quick', 'The'], key=lower)"
    )
    sd.key = str.upper
    assert sd._key == str.upper
    assert len(sd) == 5
    assert list(reversed(sd)) == ['The', 'quick', 'jumped', 'Fox', 'Brown']
    for item in sd:
        assert item in sd
    for i, item in enumerate(sd):
        assert item == sd[i]
    sd.insert('jUmPeD')
    sd.insert_right('QuIcK')
    assert sd._keys == [
        'BROWN',
        'FOX',
        'JUMPED',
        'JUMPED',
        'QUICK',
        'QUICK',
        'THE',
    ]
    assert sd._items == [
        'Brown',
        'Fox',
        'jUmPeD',
        'jumped',
        'quick',
        'QuIcK',
        'The',
    ]
    assert sd.find_le('JUMPED') == 'jumped', sd.find_le('JUMPED')
    assert sd.find_ge('JUMPED') == 'jUmPeD'
    assert sd.find_le('GOAT') == 'Fox'
    assert sd.find_ge('GOAT') == 'jUmPeD'
    assert sd.find('FOX') == 'Fox'
    assert sd[3] == 'jumped'
    assert sd[3:5] == ['jumped', 'quick']
    assert sd[-2] == 'QuIcK'
    assert sd[-4:-2] == ['jumped', 'quick']
    for i, item in enumerate(sd):
        assert sd.index(item) == i
    try:
        sd.index('xyzpdq')
    except ValueError:
        pass
    else:
        assert 0, 'Oops, failed to notify of missing value'
    sd.remove('jumped')
    assert list(sd) == ['Brown', 'Fox', 'jUmPeD', 'quick', 'QuIcK', 'The']

    # ---------------------------  Above tests are copied from SortedCollection  -------------------------
    # TODO: add test examples specific to SortedDeque
    import doctest
    from operator import itemgetter

    print(doctest.testmod())



__all__ = ['SourceReader']


from abc import ABCMeta, abstractmethod
import time

from stream2py.utility.typing_hints import (
    Optional,
    Union,
    ComparableType,
    Any,
)


class SourceReader(metaclass=ABCMeta):
    """Abstract class interface to be used by StreamBuffer.

    >>> from stream2py import SourceReader
    >>>
    >>> class SimpleCounterString(SourceReader):
    ...     '''Count in range returning a string formatted as f"s{count}"'''
    ...     def __init__(self, start, stop):
    ...         assert start < stop
    ...         self.start = start
    ...         self.stop = stop
    ...         self.range_iterator = None
    ...         self.open_count = 0
    ...
    ...     def open(self):
    ...         self.open_count += 1
    ...         self.range_iterator = iter(range(self.start, self.stop))
    ...
    ...     def read(self):
    ...         value = next(self.range_iterator, None)
    ...         if value is not None:
    ...             return f's{value}'
    ...         else:
    ...             return None
    ...
    ...     def close(self):
    ...         del self.range_iterator
    ...         self.range_iterator = None
    ...
    ...     @property
    ...     def info(self):
    ...         return dict(start=self.start, stop=self.stop, open_count=self.open_count)
    ...
    ...     def key(self, data):
    ...         return int(data[1:])
    ...
    >>> source_reader = SimpleCounterString(start=0, stop=10)
    >>> source_reader.open()
    >>> source_reader.info
    {'start': 0, 'stop': 10, 'open_count': 1}
    >>> source_reader.read()
    's0'
    >>> source_reader.close()
    >>> with source_reader:
    ...     source_reader.info
    ...     source_reader.read()
    ...     source_reader.read()
    ...     source_reader.read()
    {'start': 0, 'stop': 10, 'open_count': 2}
    's0'
    's1'
    's2'
    """

    _sleep_time_on_iter_none_s = 0.001

    def __iter__(self):
        while True:
            _next = self.read()
            if _next is not None:
                yield _next
            else:
                time.sleep(self._sleep_time_on_iter_none_s)

    def set_sleep_time_on_iter_none(self, sleep_time_s: Union[int, float] = 0.001):
        """Set the sleep time of the iter yield loop when next data item is not yet available.

        :param sleep_time_s: seconds to sleep
        """
        self._sleep_time_on_iter_none_s = sleep_time_s

    @abstractmethod
    def open(self) -> None:
        """Set up source to be read and set some source info affected by open time like the time of open.
        Will be called in StreamBuffer immediately before first read."""

    @abstractmethod
    def read(self) -> Optional[Any]:
        """Must return data that is sortable with 'key' method or None. Data that is not readily sortable such as
        ordered words of a sentence can be wrapped in a tuple (word_index, word).

        :return: data or None
        """
        raise NotImplementedError(
            "Implement the 'read' method returning data that is sortable with 'key' method"
        )

    @abstractmethod
    def close(self) -> None:
        """Close and clean up source reader.
        Will be called when StreamBuffer stops or if an exception is raised during read and append loop.
        """

    @property
    @abstractmethod
    def info(self) -> dict:
        """A dict with important source info. Default can be init_kwargs and open timestamp.

        This info will be available in BufferReader after SourceReader has opened but any changes made between open and
        close are not guaranteed to be visible. Any continuous status updates should be included with read data instead.

        :return: dict
        """
        raise NotImplementedError('no source info set')

    @abstractmethod
    def key(self, data: Any) -> ComparableType:
        """
        Converts data into a comparable value to sort by

        :param data: the return value of the 'read' method
        :return: ComparableType
        """
        raise NotImplementedError(
            "Implement the 'key' method to convert data into a comparable value to sort by"
        )

    @property
    def sleep_time_on_read_none_s(self) -> Optional[Union[int, float]]:
        """Sets default sleep time for StreamBuffer when it reads None from SourceReader.
        Useful when you know the period between data points.

        For example if data only comes once every 10 seconds,
        it would be inefficient to try and read again every 0.1 seconds and get None up to 100 times
        but every 10 seconds or even 1 second would be much more reasonable depending on the use case.
        Not necessary if read is a blocking function that always returns a data point.

        Default is None which delegates the sleep time to StreamBuffer's default.

        :return: Optional[Union[int, float]] number of seconds to sleep
        """
        return None

    timestamp_seconds_to_unit_conversion: Union[int, float] = 1e6  # to microseconds

    @classmethod
    def get_timestamp(cls) -> int:
        """Get timestamp of now as int microseconds

        :return: timestamp in UTC microseconds
        """
        return int(time.time() * cls.timestamp_seconds_to_unit_conversion)

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()
